;;; -*- Mode: Emacs-Lisp; Mode: Linkd -*-

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                                                                    ;;
;; Emacs Initialization                                               ;;
;;                                                                    ;;
;; Version:    2.4                                                    ;;
;; Author:     Christopher R. Genovese                                ;;
;; Maintainer: Christopher R. Genovese                                ;;
;;                                                                    ;;
;; Last Updated: 2010 Nov 20 Sat 21:00                                ;;
;; By:           Christopher R. Genovese                              ;;
;; Update #:     248                                                  ;;
;;                                                                    ;;
;; Copyright (C) 2010, Christopher R. Genovese, all rights reserved.  ;;                    
;;                                                                    ;;
;;                                                                    ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Commentary
;;
;; This emacs initialization is designed to support GNU Emacs 22+.
;; It has so far been tested with versions 22.2.1 and 23.2. 
;; Support for earlier emacs versions and for XEmacs is not 
;; guaranteed, although much of the functionality may work.
;;
;; This initialization file is built from a number of components
;; and scripts. If you have access to the original files, you
;; should modify those and not this one. You can access the
;; original files at
;;
;;     (@url :file-name "http://www.stat.cmu.edu/~genovese/code/emacs-init/" :display "Genovese Emacs Init")
;;
;; or email the author at (@url :file-name "mailto:genovese@cmu.edu" :display "genovese@cmu.edu").
;;
;; The intialization file works best with a number of third-party
;; and pre-installed modes. See the README file at the above
;; link for a list with version and update numbers. Most, if not
;; all, can be obtained at (@url :file-name "http://www.emacswiki.org/" :display "EmacsWiki"). The code tests for
;; the presence of these libraries, so no errors should occur
;; if they are missing. The code can be customized to incorporate
;; different or additional libraries; see the aforementioned README
;; file for instructions and see imports.el in the build directory.
;;
;; NOTE: If you have (@url :file-name "http://www.emacswiki.org/emacs/LinkdMode" :display "linkd.el"), you can engage \\[linkd-mode]
;;       to convert the table of contents below into links 
;;       throughout this file.
;;


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; TABLE OF CONTENTS

;;  1. (@> "Emacs Lisp Utility Functions")
;;  2. (@> "Environment Configuration")
;;  3. (@> "Imported Mode Requirements")
;;  4. (@> "Text Modes")
;;  5. (@> "Programming Modes")
;;  6. (@> "Shell-based Modes")
;;  7. (@> "File and Navigation Modes")
;;  8. (@> "TeX-related Modes")
;;  9. (@> "Other Modes and Features")
;; 10. (@> "ESS Mode for R and S Code")
;; 11. (@> "Global Key Bindings")
;; 12. (@> "Settings and Initialization")
;; 13. (@> "System Dependent Settings")
;; 14. (@> "Faces")
;; 15. (@> "Initial State")



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                                                           ;;
;;(@* "Emacs Lisp Utility Functions")                        ;;;;;;
;;                                                           ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;; Setting Functions for convenient adding to hooks

;(defun turn-on-auto-fill ()
;  (auto-fill-mode '1))

(defun turn-on-tex-parser-ispell ()
  (make-local-variable 'ispell-parser)
  (setq ispell-parser 'tex))

(defun turn-on-local-show-paren-mode ()
  (make-local-variable 'show-paren-mode)
  (setq show-paren-delay 0)  ; default 0.125
  (setq show-paren-style 'parenthesis)
  (show-paren-mode '1)
  (set-face-background 'show-paren-mismatch "red"))

(defun turn-on-global-show-paren-mode ()
  (setq show-paren-delay 0)  ; default 0.125
  (setq show-paren-style 'parenthesis)
  (show-paren-mode '1)
  (set-face-background 'show-paren-mismatch "red"))

(defun turn-on-local-comment-auto-fill ()
  (set (make-local-variable 'comment-auto-fill-only-comments) t)
  (auto-fill-mode '1))

(defun highlight-attn-words ()
  (font-lock-add-keywords
   nil '(("\\<\\(ATTN\\|TODO\\|FIX\\|FIXME\\|HACK\\):?"
          1 font-lock-warning-face t))))


;; Operations Functions

(defun make-backup-file-name (file)
  "Create the non-numeric backup file name for FILE.
This is a separate function so you can redefine it for customization."
  (concat (file-name-directory file) (concat "." (concat (file-name-nondirectory file) "~"))))

(defun set-abbrev-mark ()
  "Mark current point as the beginning of an abbrev.
Abbrev to be expanded starts here rather than at beginning of word before point."
  (interactive)
  (setq abbrev-start-location (point-marker)
	abbrev-start-location-buffer (current-buffer))
  (message "Abbrev Mark Set"))


;; Frame functions

(defun my-other-frame (arg)
  "Select the ARG'th different visible frame, and raise it, even if
the frame is iconified or invisible. All frames are arranged in a
cyclic order, and this command selects the frame ARG steps away in
that order.  A negative ARG moves in the opposite order."
  (interactive "p")
  (let ((frame (selected-frame)))
    (while (> arg 0)
      (setq frame (next-frame frame))
      (setq arg (1- arg)))
    (while (< arg 0)
      (setq frame (previous-frame frame))
      (setq arg (1+ arg)))
    (raise-frame frame)
    (select-frame frame)
    (set-mouse-position (selected-frame) (1- (frame-width)) 0)
    (if (fboundp 'unfocus-frame)
	(unfocus-frame))))


;; Scrolling and Cursor Movement functions

(defun scroll-down-or-beg ()
  "Smart scroll down, if too close to the beginning of the buffer,
it moves you there without error, in contrast to scroll-down."
  (interactive)
  (condition-case nil (scroll-down) (error (goto-char (point-min)))))

(defun scroll-up-or-end ()
  "Smart scroll up, if too close to the end of the buffer,
it moves you there without error, in contrast to scroll-up."
  (interactive)
  (condition-case nil (scroll-up) (error (goto-char (point-max)))))

(defun forward-sexp-or-char (&optional arg)
  "Move forward across one balanced expression (sexp) or forward
one character if none. With ARG, do it that many times.  Negative
arg -N means move backward across N balanced expressions. In
comparison to backward-sexp, this allows you to continue moving
across expressions without having to change keys."
  (interactive "p")
  (or arg (setq arg 1))
  (condition-case nil
      (forward-sexp arg)
    (error
     (let ((at-sexp-boundary? ; do we see a sexp delimiter in proper direction?
            (if (< arg 0)
                (if (char-equal (char-syntax (char-before (point))) ?\() t nil)
              (if (char-equal (char-syntax (char-after (point))) ?\)) t nil))))
       (if at-sexp-boundary?
           (forward-char arg))))))

(defun backward-sexp-or-char (&optional arg)
  "Move backward across one balanced expression (sexp) or
backward one character if at the boundary.  With ARG, do it that
many times.  Negative arg -N means move forward across N balanced
expressions.  In comparison to backward-sexp, this allows you to
continue moving across expressions without having to change
keys."
  (interactive "p")
  (or arg (setq arg 1))
  (forward-sexp-or-char (- arg)))

(defun my-move-beginning-of-line (&optional n)
  "If not at the beginning of a line, move point to the beginning
of the line, moving forward n - 1 lines first if n is not nil or
1. However, if at the beginning of a line, move back to the first
non-ws character on the line."
  (interactive "p")
  (if (looking-at "^")
      (back-to-indentation)
    (move-beginning-of-line n)))

;; Region and Buffer Functions

  ;; I like a message even if the entire region is visible
  ;; and transient mark mode is on. Because it is conditional
  ;; cannot implement this easily with advice.

(defun my-kill-ring-save (beg end)
  "Save the region as if killed, but don't kill it.
In Transient Mark mode, deactivate the mark.
If `interprogram-cut-function' is non-nil, also save the text for a window
system cut and paste.

If you want to append the killed line to the last killed text,
use \\[append-next-kill] before \\[kill-ring-save].

This command is similar to `copy-region-as-kill', except that it gives
visual feedback indicating the extent of the region being copied, and
prints a message to the echo area accordingly."
  (interactive "r")
  (copy-region-as-kill beg end)
  ;; This use of interactive-p is correct
  ;; because the code it controls just gives the user visual feedback.
  (if (interactive-p)
      (let ((other-end (if (= (point) beg) end beg))
	    (opoint (point))
	    ;; Inhibit quitting so we can make a quit here
	    ;; look like a C-g typed as a command.
	    (inhibit-quit t))
	(if (pos-visible-in-window-p other-end (selected-window))
	    (if (and transient-mark-mode
                     (face-background 'region))
                (message "Region Copied")
	      ;; Swap point and mark.
	      (set-marker (mark-marker) (point) (current-buffer))
	      (goto-char other-end)
	      (sit-for blink-matching-delay)
	      ;; Swap back.
	      (set-marker (mark-marker) other-end (current-buffer))
	      (goto-char opoint)
	      ;; If user quit, deactivate the mark
	      ;; as C-g would as a command.
	      (and quit-flag mark-active
		   (deactivate-mark)))
	  (let* ((killed-text (current-kill 0))
		 (message-len (min (length killed-text) 40)))
	    (if (= (point) beg)
		;; Don't say "killed"; that is misleading.
		(message "Saved text until \"%s\""
			(substring killed-text (- message-len)))
	      (message "Saved text from \"%s\""
		      (substring killed-text 0 message-len))))))))

  ;; Standard version is silent, so it is hard to tell if the key was hit
  ;; Only added a message. ATTN: This could be implemented with advice,
  ;; but copy-region-as-kill is used programmatically in other functions
  ;; (e.g., kill-ring-save) that may not want a message printed.

(defun my-copy-region-as-kill (beg end)
  "Save the region as if killed, but don't kill it.
If `interprogram-cut-function' is non-nil, also save the text for a window
system cut and paste.  Produces a message to confirm copy."
  (interactive "r")
  (if (eq last-command 'kill-region)
      (kill-append (buffer-substring beg end) (< end beg))
    (kill-new (buffer-substring beg end)))
  (message "Region Copied")
  (if transient-mark-mode
      (setq deactivate-mark t))
  nil)

(defun eval-and-replace-sexp ()
  "Replace the preceding emac-lisp sexp in current buffer
   with its value."
  (interactive)
  (backward-kill-sexp)
  (condition-case nil
      (prin1 (eval (read (current-kill 0)))
             (current-buffer))
    (error (message "Invalid expression")
           (insert (current-kill 0)))))


;; Import and Import-Conditional Execution

;; ATTN: macro version not appropriate because we need to 'apply this
;(defmacro require-soft (feature &optional file)
;  "Try to require FEATURE, but don't signal an error if `require' fails."
;  `(require ,feature ,file 'noerror)
;  )

(defun require-soft (feature &optional file)
  "Try to require FEATURE, but don't signal an error if `require' fails."
  (require feature file 'noerror))

(defmacro with-library (symbol &rest body)
   "Execute BODY conditionally on the availability of library SYMBOL"
   `(if (not (require ,symbol nil t))
        (message (format "Library %s is not available." ',symbol))
      ,@body))


;; List Functions

(defun remove-matching-elements (elt lst &optional pred)
  "Remove elements in LST matching ELT according to the
predicate PRED. If PRED is not supplied, it defaults to equal."
  (let ((pred (or pred (symbol-function 'equal)))
	(newlst nil)
	cmp)
    (while lst
      (setq cmp (car lst))
      (if (not (funcall pred elt cmp))
	  (setq newlst (cons cmp newlst)))
      (setq lst (cdr lst)))
    (reverse newlst)))
      
(defun remove-non-matching-elements (elt lst &optional pred)
  "Remove elements in LST matching ELT according to the
predicate PRED. If PRED is not supplied, it defaults to equal."
  (let ((pred (or pred (symbol-function 'equal)))
	(newlst nil)
	cmp)
    (while lst
      (setq cmp (car lst))
      (if (funcall pred elt cmp)
	  (setq newlst (cons cmp newlst)))
      (setq lst (cdr lst)))
    (reverse newlst)))

(defun concatenate-list (&rest seqs)
  "Concatenate arguments into a flattened list."
  (apply 'append (append seqs '(nil))))


;; Predicates for element removal
      
(defun eq-car (l1 l2)
  "Returns t if car of l1 and l2 are eq, nil otherwise.
If l1 (or l2) are not cons cells, l1 (or l2) itself is used in the comparison.
This is useful as a predicate for removing elements from an association list;
see remove-matching-elements.  In this case, if the keys are symbols then
only the symbol of interest need be passed to remove-matching elements."
  (eq (if (consp l1) (car l1) l1) (if (consp l2) (car l2) l2)))

(defun equal-car (l1 l2)
  "Returns t if car of l1 and l2 are equal, nil otherwise.
If l1 (or l2) are not cons cells, l1 (or l2) itself is used in the comparison.
This is useful as a predicate for removing elements from an association list;
see remove-matching-elements.  In this case, if the keys are symbols then
only the symbol of interest need be passed to remove-matching elements."
  (equal (if (consp l1) (car l1) l1) (if (consp l2) (car l2) l2)))


;; File and Directory Name Functions and Predicates

(defun expand-directory-name (dir)
  "Expands a directory name as in expand-file-name but ensures
that the name ends with a '/'.  This can be used to convert
a directory name, even in relative or ~ form, to a canonical
form, for comparison, concatenation, or other purposes."
  (save-excursion
    (let ((dir (and (stringp dir) (expand-file-name dir))))
      (if (and dir (not (string-match "/$" dir)))
	  (concat dir "/")
	dir))) )

(defun equal-file-name (fn1 fn2 &optional func)
  "Tests whether two file-names FN1 and FN2 refer to the same file.
This uses FUNC to bring both file-names to a common form (e.g., abolute).
If FUNC is not provided, it defaults to a expand-file-name."
  (if (and (stringp fn1) (stringp fn2))
      (let ((func (or func (symbol-function 'expand-file-name))))
	(string-equal (funcall func fn1) (funcall func fn2)))))

(defun equal-dir-name (fn1 fn2 &optional func)
  "Tests whether two file-names FN1 and FN2 refer to the same file.
This uses FUNC to bring both file-names to a common form (e.g., abolute).
If FUNC is not provided, it defaults to a function that does
expand-file-name followed by appending a trailing slash."
  (if (and (stringp fn1) (stringp fn2))
      (let ((func (or func (symbol-function 'expand-directory-name))))
	(string-equal (funcall func fn1) (funcall func fn2)))))

;; Macros

(defmacro do-only-once (loadvar &rest body)
  "Do BODY if LOADVAR is nil and set LOADVAR to t.
If LOADVAR is not nil, do nothing. LOADVAR can be
an arbitrary place, as accepted by setf. It should
be NIL initially."
  (declare (indent 1) (debug t))
  `(if ,loadvar
       nil
     ,@body
     (setf ,loadvar t)))


  ;; Adapted for elisp from Paul Graham's "ANSI Common Lisp"
(defmacro my-aif (test true-body &rest false-body)
  "Evaluate TRUE-BODY or FALSE-BODY depending on value of TEST.
If TEST returns non-nil, bind `it' to the value, and evaluate
TRUE-BODY.  Otherwise, evaluate forms in FALSE-BODY as if in `progn'.
Compare with `if'."
  (declare (indent 2))
  (let ((sym (make-symbol "my-tmp-aif-sym7814")))
    `(let ((,sym ,test))
       (if ,sym
	   (let ((it ,sym))
	     ,true-body)
	 (progn
	   ,@false-body)))))

(defmacro my-awhen (test &rest body)
  "Evaluate BODY if TEST returns non-nil.
During evaluation of body, bind `it' to the value returned by TEST."
  (declare (indent 1))
  `(my-aif ,test
       (progn ,@body)
     nil))


;; Miscellaneous Useful Functions and Tools

(defun my-get-makefile-targets ()
  "Scan the contents of all files makefiles in the current
directory and return a list of targets.  The files scanned match
the glob [mM]akefile* within .  All targets are returned except
those containing $ (for variable substitution) or % for pattern
matching."
  (with-temp-buffer
    (let ((makefiles (nreverse (file-expand-wildcards "[mM]akefile*")))
          targets)
      (dolist (file makefiles)
        (insert-file-contents (concat default-directory file)))
      (goto-char (point-min))
      ; Simply exclude $ and %. Can add more generality later if needed.
      (while (re-search-forward "^\\s-*\\([^$%: 	\n]+?\\)\\s-*:" nil t)
        (setq targets (cons (match-string 1) targets)))
      (nreverse (cons "" targets)))))


(defun view-url-in-buffer ()
  "Open a new buffer containing the contents of URL. If thingatpt+
   is loaded, tries to find URL nearest point as default; otherwise,
   no default is given."
  (interactive)
  (let* ((default (if (featurep 'thingatpt+) (thing-nearest-point 'url) "")) 
         (url (read-from-minibuffer "URL: " default))
         (urlbuf (url-retrieve-synchronously url)))
    (when urlbuf
      (switch-to-buffer urlbuf)
      (rename-buffer url t)
      (goto-char (point-min))
      ;; For an http request, header is included in the buffer,
      ;; so remove up to and including first blank line
      (when (and (looking-at "HTTP")
                 (re-search-forward "^\\s-*$" nil t))
        (condition-case nil (forward-char 1) (error nil)) ; move past blank line
        (delete-region (point-min) (point)))
      (setq case-fold-search t) ; buffer local
      ;; This should use nxml-mode or nxhtml-mode automatically
      (cond ((looking-at "<?xml") (xml-mode))
            ((looking-at "<\\(!DOCTYPE\\s-+\\)html") (html-mode))))))


(defun new-buffer (&optional noselect)
  "Create a new buffer named *Untitled*<n> for the n that makes it unique.
If NOSELECT is nil, switch to that buffer; otherwise just display it."
  (interactive "P")
  (let ((buf (get-buffer-create (generate-new-buffer-name "*Untitled*"))))
        (if (null noselect)
            (switch-to-buffer buf)
          (display-buffer buf))))
          ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                                                           ;;
;;(@* "Environment Configuration")                           ;;;;;;
;;                                                           ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; This component is automatically generated from a template
;; by ./makeconfig, perl, and make. See Makefile for details.


(defvar my-user-name "genovese"
  "*If non-nil, this should be set to a string giving the user name
to use by default in locating the ....  If set to nil, this will
always use the HOME directory in the environment if it is available
as the root for the startup information.  Set this is you wish to
load your startup file when running as someone else via the -u flag.")

(defvar my-home-aliases (list "/Users/genovese/")
  "*This is a list of strings indicating possible aliases for the users
HOME directory that emacs might not recognize.  The first element of the list should be the
prefered or most common or most specific form.  This variable is most useful with
cross-mounted file systems, and is used by functions to determine whether 
or not a directory name should be changed to \"~/\".")

(defvar my-home-dir (if (and (equal my-user-name (getenv "USER")) (getenv "HOME"))
                        (concat (getenv "HOME") "/")
                      (or  (car my-home-aliases) ""))
  "Home directory")

(defvar my-site-lisp-dir (expand-directory-name (if (= emacs-major-version 22) "/Applications/Emacs22.app/Contents/Resources/site-lisp/" "/Applications/Emacs.app/Contents/Resources/site-lisp/"))
  "The site-lisp directory (absolute path) containing the globally available
   add-on modes and .el files. This name must end in /.")

(defvar my-home-lisp-dir (concat "~/.emacs.d/"
                                 (if (null (string-match "/$" "~/.emacs.d/")) "/"))
  "The site-lisp directory (absolute path) containing the users local
   add-on modes and .el files. This name must end in /.")

(when (>= emacs-major-version 23)
  (setq user-emacs-directory my-home-lisp-dir))

(defvar my-frame-alist `((top . 1) (left . 1) (width . ,(if (= emacs-major-version 23) 224 180)) (height . ,(if (= emacs-major-version 23)  64  62)))
  "Geometry parameters that will be used for initial-frame-alist to determine initial window size.")

(if (and (boundp 'my-home-dir) (stringp my-home-dir)
         (not (string= my-home-dir "")) (not (member my-home-dir my-home-aliases)))
    (setq my-home-aliases (cons my-home-dir my-home-aliases)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                                                           ;;
;;(@* "Imported Mode Requirements")                          ;;;;;;
;;                                                           ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun import (feature-list &optional action &rest xargs)
  "Apply ACTION to each of features given in FEATURE-LIST,
   supplying additional arguments if necessary. Typical
   use case is for ACTION to be require, require-soft,
   or autoload. If not supplied, ACTION defaults to
   require.

   Elements of FEATURE-LIST can have several forms.
   The feature can one of the following forms:

       1. a SYMBOL,
       2. a cons (CONDITION . SYMBOL),
       3. a list (SYMBOL ARGS...), or
       4. a list (CONDITION SYMBOL ARGS...).

   Here, SYMBOL is a symbol corresponding to a feature,
   typically one that can be validly require'd.
   And CONDITION is an expression that eval's
   to a boolean (validly at load and compile time)
   and is used to selectively load particular features.
   A typical example would be (>= emacs-major-version 22).
   ARGS is a list of 0 or more additional arguments.
   (In cases 1 and 2 above, ARGS will effectively be an
   an empty list.) The lists ARGS and that supplied in
   the &rest argument XARGS will be concatenated, in
   that order, and the corresponding elements will be
   passed as arguments to the ACTION function."
  (let ((req (or action 'require)))
    (dolist (feature feature-list)
      (let (sym args)
        (cond
         ;; Feature is a symbol
         ((symbolp feature)
          (setq sym feature)
          (setq args xargs)
          )
         ;; Feature is (list SYMBOL ARGS...)
         ((and (consp feature)
               (symbolp (car feature)))
          (setq sym (car feature))
          (setq args (concatenate-list (cdr feature) xargs))
          )
         ;; Feature is (CONDITION . SYMBOL) or (list CONDITION SYMBOL ARGS...)
         ((and (consp feature)
               (booleanp (eval (car feature)))
               (or (symbolp (cdr feature))
                   (and (listp (cdr feature))
                        (symbolp (cadr feature)))) )
          (when (eval (car feature))
            (if (symbolp (cdr feature))  ;; have cons (CONDITION . SYMBOL)
                  (progn
                    (setq sym (cdr feature))
                    (setq args xargs)
                    )                    ;; else have (list CONDITION SYMBOL ...)
              (setq sym (cadr feature))
              (setq args (concatenate 'list (cddr feature) xargs))
              ))
          )
         (t
          (message (concat "Erroneous feature specification: " (prin1-to-string feature)))
          ))
        (if sym (apply req sym args))
        ))
    ))

(defalias 'deferred-import 'identity
  "Syntactic sugar to allow user to saliently list
   those modules that will be imported later in the
   init file but not in the main import list. This
   is usually done for timing purposes, for instance
   to ensure that hooks are properly defined at
   load time.")


;;
;; Lists of Packages to Import
;;

(defvar my-must-features
  '(comint shell calc)
  "*List of *built-in* features that must be required for later
   processing in this initialization file. Each element is either
   a symbol that is a valid argument to require or is a cons-cell
   whose car is a major version number for Emacs and whose cdr is
   a symbol that is a valid argument to require. These features
   will be required, and failure of the require will lead to an
   error.")

(defvar my-ought-features
  '(org-install icicles ido dired-x dired-aux)
  "*List of features that are distributed with emacs that would be
   desireable to require for later processing in this initi file. 
   If for some reason, they are not available, a message will be
   displayed but no error produced.

   Each element of the list is either a symbol that is a valid
   argument to require or is a cons-cell whose car is a major
   version number for Emacs and whose cdr is a symbol that is a
   valid argument to require.")

(defvar my-want-features
  '(dired+ bookmark+ linkd lacarte dropdown-list
           yasnippet anything-config arduino-mode)
  "*List of features, not necessarily available, that it would be 
   desireable to require for later processing in this init file. 
   If for some reason, they are not available, a message will be
   displayed but no error produced.

   Each element of the list is either a symbol that is a valid
   argument to require or is a cons-cell whose car is a major
   version number for Emacs and whose cdr is a symbol that is a
   valid argument to require.")

(defvar my-autoloaded-features
  (if (equal emacs-major-version 22)
      (list '(nxml-mode   "nxml/autostart.el" "nXML Mode"   t)
            '(nxhtml-mode "nxml/autostart.el" "nXHTML Mode" t)
            '(tuareg-mode "tuareg" "Major mode for editing Caml code" t)
            '(camldebug "camldebug" "Run the Caml debugger" t)
            '(js2-mode "js2" "JavaScript Mode" t)
            '(texmathp-compile "texmathp" "TeX Math Parsing" t)
            )
    (list ))
  "*List of features to be autoloaded along with the arguments
   to be passed to autoload. Elements of the lists are of
   the form (SYMBOL ARGS...).")

(defvar my-special-features
  (list
   (list 'ibuf-ext   '(deferred-import "Only loaded as needed when ibuffer is first instantiated"))
   (list 'texmathp   '(deferred-import "Only loaded as needed when AucTeX is first instantiated"))
   (list 'ess-site   '(deferred-import "Temporarily deferring load to ess.el for timing purposes"))
   (list 'thingatpt+ '(eval-after-load "thingatpt"
                        '(require-soft 'thingatpt+)))
   (list 'cedet      '(if (not (featurep 'cedet))
                            (load-file (concat my-site-lisp-dir "cedet/common/cedet.el"))))
   (list 'jde        '(progn
                        (setq defer-loading-jde t)
                        (if (not defer-loading-jde)
                            (require-soft 'jde)
                          (autoload 'jde-mode "jde" "JDE mode." t)
                          (add-to-list 'auto-mode-alist '("\\.java\\'" . jde-mode))
                          )))
   (list 'ecb        '(progn
                        (eval-when-compile
                          (if (not (featurep 'cedet))
                              (require 'cedet)))
                        (require-soft 'ecb)
                        ))
   )
  "List of features that will need special processing (usually beyond
   a simple require) for use in this init file. Although somewhat
   awkward, the code to handle these cases should be given as
   a form in the list. Specifically, each element of the list
   should in turn be a list with either two or three elements
   corresponding to 

      (list 'SYMBOL 'BODYFORM) 
      (list  CONDITION 'SYMBOL 'BODYFORM),

   where CONDITION is a lisp expression that evaluates to t
   if this feature should be acted upon. A typical example
   for condition would be (>= emacs-major-version 22).
   In any case, CONDITION must be valid for evaluation at
   either load or compile time.")


;;
;; Do Imports and Autoloads
;;

(import my-must-features       'require)
(import my-ought-features      'require-soft)
(import my-want-features       'require-soft)
(import my-autoloaded-features 'autoload)
(import my-special-features    (lambda (sym body) (eval body)))


;;
;; Automatic Mode Detection
;;

(add-to-list 'auto-mode-alist '("\\.org\\'" . org-mode))
(add-to-list 'auto-mode-alist '("\\.txt\\'" . org-mode))
(add-to-list 'auto-mode-alist '("\\.[0-9]\\'" . fundamental-mode))
(add-to-list 'auto-mode-alist '("\\.[sS]\\'" . S-mode))
(add-to-list 'auto-mode-alist '("\\.[rR]\\'" . R-mode))
(add-to-list 'auto-mode-alist '("\\.\\([pP]\\([LlmM]?\\|od\\|erl\\)\\|al\\)\\'" . perl-mode))
(add-to-list 'auto-mode-alist 
             (cons (concat "\\." (regexp-opt '("xml" "xsd" "sch" "rng" "xslt" "svg" "rss") t) "\\'")
                   'nxml-mode))
(add-to-list 'auto-mode-alist '("\\.s?html?\\'" . nxhtml-mode))
(add-to-list 'auto-mode-alist '("\\.js\\'" . js2-mode))
(add-to-list 'auto-mode-alist '("\\.ml[iylp]?\\'" . tuareg-mode))
(when (null (featurep 'arduino-mode))
  (add-to-list 'auto-mode-alist '("\\.pde\\'" . c-mode)))


(fset 'xml-mode 'nxml-mode)    ;; should rhs be (symbol-function 'nxml-mode) ??
(fset 'html-mode 'nxhtml-mode)

(add-to-list 'magic-mode-alist '("<\\?xml " . nxml-mode))
(add-to-list 'magic-mode-alist '("#!.*perl" . perl-mode))
(add-to-list 'magic-mode-alist '("#!.*python" . python-mode))
(add-to-list 'magic-mode-alist '("#!.*\\(t?c\\|ba\\|k\\|z\\|\\)sh" . shell-script-mode))
(add-to-list 'magic-mode-alist '("\\\\input " . plain-TeX-mode))
(add-to-list 'magic-mode-alist '("\\\\document\\(style\\|class\\)" . LaTeX-mode))


;;
;; Common Lisp facilities (only used here and there)
;;

(eval-when-compile (require 'cl))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                                                           ;;
;;(@* "Text Modes")                                          ;;;;;;
;;                                                           ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Generic Text Modes

(defun my-generic-text-modes-hook ()
  "Sets up environment useful for a variety of text editing modes.
   Add this to the corresponding hooks."
  (turn-on-flyspell)
  (turn-on-local-comment-auto-fill))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Org Mode

(defun my-org-load-hook ()
  (setq org-startup-folded nil)
  )

(defun my-org-mode-hook ()
  (local-set-key "\C-c\C-x\M-k" 'org-cut-special)
  (local-set-key "\C-c\C-x\M-c" 'org-copy-special)
  (local-set-key "\C-c\M-c"     'org-edit-src-code) ; see which I like best
  (local-set-key "\C-c\M-s"     'org-edit-src-code) ; C-c ' taken by icicles
  )

(add-hook 'org-load-hook 'my-org-load-hook)
(add-hook 'org-mode-hook 'turn-on-font-lock)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                                                           ;;
;;(@* "Programming Modes")                                   ;;;;;;
;;                                                           ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Generic Programming Modes

(defun my-generic-prog-modes-hook ()
  "Sets up environment useful for a variety of programming modes.
   Add this to the corresponding hooks."
  (highlight-attn-words)
  (turn-on-local-comment-auto-fill))

(when (featurep 'header2)
  (setq header-copyright-notice "Copyright (C) 2010, Christopher R. Genovese, all rights reserved.\n")
  (setq header-date-format "%Y %b %d %a %H:%M:%S"))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; IDE Configuration CEDET/ECB/JDEE  (necessary files loaded in imports.el)

(defun my-semantic-init-hook ()
  (global-semantic-idle-scheduler-mode '1)
  (semantic-load-enable-excessive-code-helpers)
  (require 'semantic-ia))

(add-hook 'semantic-init-hook 'my-semantic-init-hook)

(eval-after-load 'senator   ;; alternatively could put this in a hook
  '(progn
     (define-key senator-mode-map "\C-c,c"    'semantic-ia-describe-class)
     (define-key senator-mode-map "\C-c,d"    'semantic-ia-show-doc)
     (define-key senator-mode-map "\C-c,e"    'eassist-list-methods)
     (define-key senator-mode-map "\C-c,F"    'semantic-ia-fast-jump)
     (define-key senator-mode-map "\C-c,s"    'semantic-symref)
     (define-key senator-mode-map "\C-c,S"    'semantic-symref-symbol)
     (define-key senator-mode-map "\C-c,x"    'semantic-symref-regexp)
     (define-key senator-mode-map "\C-c,\C-k" 'senator-kill-tag)
     (define-key senator-mode-map "\C-c,\M-c" 'senator-copy-tag)
     (define-key senator-mode-map "\C-c,\C-w" nil)
     (define-key senator-mode-map "\C-c,\M-w" nil)
     ))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Emacs-Lisp Mode

(defun my-emacs-lisp-mode-hook ()
  (local-set-key "\C-c\C-e" 'eval-defun)
  (local-set-key "\C-c\C-o" 'outline-minor-mode)
  (local-set-key "\C-c\C-s" 'eval-last-sexp)
  (defun emacs-lisp-outline-minor-setup ()
    (setq outline-regexp ";;; \\|;; \\|(....")
    (local-set-key "\C-c\C-m" outline-mode-prefix-map)
    (when (featurep 'org)
      (define-key outline-minor-mode-map [(control tab)] 'org-cycle)
      (define-key outline-minor-mode-map [(shift tab)]   'org-global-cycle))) 
  (make-local-variable 'outline-minor-mode-hook)
  (add-hook 'outline-minor-mode-hook 'emacs-lisp-outline-minor-setup))

(add-hook 'emacs-lisp-mode-hook 'highlight-attn-words)
(add-hook 'emacs-lisp-mode-hook 'my-emacs-lisp-mode-hook)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Python Mode

(defun my-python-mode-hook ()
  (local-set-key "\M-\C-a"  'beginning-of-python-def-or-class)
  )

(add-hook 'python-mode-hook 'my-python-mode-hook)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Common C Modes (cc-mode) (C, C++, Java)

(autoload 'c-mode    "cc-mode" "C Editing Mode" t)    ;; Use cc-mode for all C editing
(autoload 'c++-mode  "cc-mode" "C++ Editing Mode" t)
(autoload 'java-mode "cc-mode" "Java Editing Mode" t)

(defun my-c-mode-common-hook ()
  (if (not (assoc "CRG" c-style-alist))
      (c-add-style "CRG" (cdr c-crg-style)))
  (c-set-style "CRG")
  (if (assq 'c-auto-hungry-string minor-mode-alist)   ;; Inhibit minor mode notation
      (setq minor-mode-alist
	    (cons '(c-auto-hungry-string "") (remove-matching-elements 'c-auto-hungry-string minor-mode-alist 'eq-car))))
  (local-set-key "\C-c\C-a" 'c-beginning-of-statement)
  (local-set-key "\C-c\C-e" 'c-end-of-statement)
  (local-set-key "\C-c\C-b" 'c-backward-conditional)
  (local-set-key "\C-c\C-f" 'c-forward-conditional)
  (local-set-key "\C-c\C-c" 'comment-region)
  (local-set-key "\M-q"     'c-fill-paragraph)
  (local-set-key "\C-c\C-d" 'c-down-conditional)
  (local-set-key "\C-c\C-u" 'c-up-conditional)
  (local-set-key "\C-c\C-i" 'c-indent-exp)
  (local-set-key "\C-c\C-q" 'c-indent-defun)
  (local-set-key "\C-c\C-m" 'compile);; m for make
  (local-set-key "\C-c\C-n" 'up-list)
  (local-set-key "\C-c\C-p" 'c-up-list-neg)
  (local-set-key "\C-c\C-o" 'outline-minor-mode)
  (local-set-key "\C-c\C-t" 'c-set-offset)
  (local-set-key "\C-c\C-x" 'c-macro-expand)
  (local-set-key [f4]       'c-fill-paragraph)
  )
  
(add-hook 'c-mode-common-hook 'highlight-attn-words)
(add-hook 'c-mode-common-hook 'my-c-mode-common-hook)

(defvar c-crg-style
  '("CRG"
    (c-basic-offset . 4)
    (c-comment-only-line-offset . (0 . 0))
    (c-offsets-alist . ((statement-block-intro . +)
			(substatement-open . 0)
			(label . 0)
			(case-label . *)
			(access-label . /)
			(statement-case-intro . *)
			(statement-cont . +)
			(cpp-macro . 0)
                        (inline-open . 0)
			(inher-intro . c-lineup-inher-intro)
			))
    (c-auto-newline . t)
    (c-hanging-braces-alist . ((brace-list-open)
			       (brace-list-close)
			       (substatement-open before after)
			       (block-close before after)))
    (c-hanging-comment-ender-p . nil)
    (c-cleanup-list . '(scope-operator list-close-comma))
    )
  "My personal formatting style for editing C/C++/Java source.")

(defun c-up-list-neg (arg)
  "Like up-list but with a negated arg"
  (interactive "P")
  (if (not arg)
      (up-list -1)
    (up-list arg)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; C-mode (old)

(defun my-c-mode-hook ()
  (setq indent-tabs-mode nil)
  (local-set-key "\C-c\C-z" 'compilation-close-compilation)
  )

(add-hook 'c-mode-hook 'my-c-mode-hook)

(defvar boc-mode-hook                    
  '((lambda ()
      (setq c-indent-level '4)
      (setq c-continued-statement-offset '4)
      (setq c-continued-brace-offset '-4)
      (setq c-brace-offset '0)
      (setq c-brace-imaginary-offset '0)
      (setq indent-tabs-mode 'nil) 
      (setq c-label-offset '-2)
      (setq c-auto-newline t)
      ))
  "Style hook for \"Boring Old C-mode\" in case
it is needed to run an earlier version of emacs.")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; C++-mode

(defun my-c++-mode-hook ()
  (setq indent-tabs-mode nil)
  (local-set-key "\C-c\C-z" 'compilation-close-compilation)
  )

(add-hook 'c++-mode-hook 'my-c++-mode-hook)

(defun c-lineup-inher-intro (langelem)
  "Line up inheritence intro : two characters past class keyword"
  (+ c-basic-offset (/ c-basic-offset 2))) 

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; java-mode

(defun my-java-mode-hook ()
  (setq indent-tabs-mode nil)
  )

(add-hook 'java-mode-hook 'my-java-mode-hook)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; compilation-mode

(defun my-compilation-mode-hook ()
  (local-set-key "\C-c\C-z" 'compilation-close-compilation)
  )

(add-hook 'compilation-mode-hook 'my-compilation-mode-hook)

(defun compilation-close-compilation (arg)
  "Iconifies compilation pop-up frame.  With prefix arg,
kills *compilation* buffer and thus its dedicated frame,
if the latter exists."
  (interactive "P")
  (let* (err-msg buf)
    (setq buf (condition-case err-msg
		  (setq buf (compilation-find-buffer))
		(error (and (message (car (cdr err-msg))) nil))))
    (if (and buf (not arg))
	(let ((win (get-buffer-window buf t)))
	  (if win
	      (iconify-frame (window-frame win))))
      (if (and buf arg)
	  (kill-buffer buf))
      )))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; perl-mode


(defun my-perl-mode-hook ()
  (setq perl-indent-level '4)
  (setq perl-continued-statement-offset '4)
  (setq perl-continued-brace-offset '-4)
  (setq perl-brace-offset '0)
  (setq perl-brace-imaginary-offset '0)
  (setq indent-tabs-mode 'nil) 
  (setq perl-label-offset '-2)
  )

(add-hook 'perl-mode-hook 'highlight-attn-words)
(add-hook 'perl-mode-hook 'my-perl-mode-hook)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; OCaml tuareg-mode

    ;; no hook yet...

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                                                           ;;
;;(@* "Shell-based Modes")                                   ;;;;;;
;;                                                           ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; ipython

;(require-soft 'ipython)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; comint-mode 

;; (autoload 'my-comint-bol "my-comint-ext")  ;; ATTN: I think this is right to remove

(defun my-comint-mode-hook ()
  (setq comint-input-ring-size '1024)
  (setq comint-input-ignoredups t)
  (local-set-key [?\C-c return] 'comint-copy-old-input)
  (local-set-key "\C-c\C-m"    nil)  ;; used globally for extended command
  (add-hook 'comint-output-filter-functions 'comint-strip-ctrl-m)
  )
       
(add-hook 'comint-mode-hook 'my-comint-mode-hook)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Extension to comint-mode, provides a few convenient functions
;;;

(require 'comint)

(defun my-comint-bol ()
  "Goes to the beginning of line, then skips past the prompt, if any.
However, if already at the prompt, the prompt is not skipped.
This has the side-effect of skipping the prompt if you start at the
beginning of a line with a prompt.  This oscillation between the two
possible beginnings is useful.

The prompt skip is done by skipping text matching the regular expression
`comint-prompt-regexp', a buffer local variable."
  (interactive)
  (let ((now (point))
	(eol (save-excursion (end-of-line) (point))))
    ;; (beginning-of-line)  ;; ATTN: this got confused by "field boundaries" in later versions of emacs
    (forward-line 0)
    (if (looking-at comint-prompt-regexp)
	(if (and (not (= (match-end 0) now))
		 (<= (match-end 0) eol))
	    (goto-char (match-end 0))))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; shell-mode

(setq-default shell-file-name "/bin/tcsh")

(defun my-shell-mode-hook ()
  ;; (require 'my-shell-ext)
  (local-set-key "\C-c\C-q" 'shell-toggle-autoexpand)
  (local-set-key "\C-a" 'my-comint-bol)
  (setq shell-pushd-regexp "\\(pd\\|cd\\|pushd\\)")
  (setq shell-cd-regexp    "chdir")
  (setq shell-popd-regexp  "\\(pf\\|popd\\)")
  (setq shell-prompt-pattern "^[^#%$>]*[#$%>]>? *" )
  (setq shell-pushd-dextract t)
  (setq shell-pushd-dunique t)
  (setq shell-pushd-silent nil)
  (setq shell-pushd-tohome t)
  (setq comint-process-echoes t)
  (setq comint-input-autoexpand 'input)
  (setq shell-input-autoexpand 'input)
  (setq shell-backup-input-autoexpand (cons nil shell-input-autoexpand))
  ;;
  ;; Check that starting directory is correct on startup
  ;;
  (setq default-directory (shell-map-home-aliases default-directory))
  ;;
  ;; Include cwd on the mode line
  ;;
  (make-local-variable 'mode-line-format)
  (setq mode-line-format
        (reverse (append
                  (list " -%-" 'default-directory "-- ")
                  (cdr (reverse (default-value 'mode-line-format))))))
  ;;
  ;; Configure Output Filter 
  ;;
  (add-hook 'comint-output-filter-functions 'comint-watch-for-password-prompt nil t)
  ;;
  ;; Set-up replacements for shell.el functions, mine given below
  ;;
  (use-my-shell-ext)
  )

(add-hook 'shell-mode-hook 'my-shell-mode-hook)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; Extension to shell-mode to improve the directory stack handling
;; and to add a few convenient features.
;;

(require 'shell)

(defvar shell-pushd-silent nil
  "*If non-nil, show dirstack in minibuffer with every change
of the directory stack.  This mimicks the optional behavior
of tcsh.")

(defvar shell-backup-input-autoexpand (cons nil 'input)
  "*List of autoexpand states among which the user can toggle with
shell-toggle-autoexpand.")

(defvar shell-home-aliases (if (boundp 'my-home-aliases)
                               my-home-aliases
                             (if (getenv "HOME")
                                 (list (concat (getenv "HOME") "/"))
                               nil))
  "*This is a list of strings indicating possible aliases for the users
HOME directory that emacs might not recognize.  This is most useful with
cross-mounted file systems, and is used by shell-map-home-aliases to 
determine whether or not the directory should be changed to \"~/\".")


(defun use-my-shell-ext ()
  "Set selected shell mode functions to my extended versions."
  (fset 'shell-resync-dirs
        (symbol-function 'my-shell-resync-dirs))
  (fset 'shell-directory-tracker
        (symbol-function 'my-shell-directory-tracker))
  (fset 'shell-process-cd
        (symbol-function 'my-shell-process-cd))
  (fset 'shell-process-pushd
        (symbol-function 'my-shell-process-pushd))
  (fset 'shell-dirstack-message
        (symbol-function 'my-shell-dirstack-message))
  (fset 'shell-replace-by-expanded-directory
        (symbol-function 'my-shell-replace-by-expanded-directory)) )

(defun shell-toggle-autoexpand ()
 "Toggle history auto-expansion to eliminate strange behavior in interactive
  applications run within the shell.  The list shell-backup-input-autoexpand
  stores the list of values to be toggled among.  At the moment, don't set
  shell/comint-input-autoexpand without changing this list as well."
  (interactive)
  (let ((nextia  (car shell-backup-input-autoexpand))
        (otheria (cdr shell-backup-input-autoexpand))
        (curia   comint-input-autoexpand))
    (setq comint-input-autoexpand nextia)
    (setq shell-backup-input-autoexpand (cons otheria nextia))
    (if nextia
       (message (format "Setting Input Autoexpand to %s" nextia))
       (message "Turning Off Input Autoexpand"))))

 ;;; Additions and fixes to shell directory tracking mechanism,
 ;;; based on the code in shell.el Gnu Emacs 19.30.1.
 ;;;
 ;;; Features:
 ;;; 
 ;;; 1. Support =n and =- references to directory stack (without expansion) as in tcsh
 ;;; 2. Fixed expansion of =n and =- (when instigated) to prevent missing slash
 ;;; 3. Sets shell-last-dir properly so  cd - works as in tcsh
 ;;; 4. Fixed bug in shell-resync-dirs in case dir tracking gets lost
 ;;; 5. Improved handling of aliases for HOME directory.


(defun shell-map-home-aliases (indir &optional noslash)
  "Check if DIR matches one of the aliases for HOME
and thus should be replaced by \"~/\" for consistency."
  (let ((dir indir))
    (if (not (string-match "/$" dir))
	(setq dir (concat dir "/")))
    (if (and shell-home-aliases (member dir shell-home-aliases))
	"~/"
      (if noslash indir dir)) ))

(defun my-shell-resync-dirs ()
  "Resync the buffer's idea of the current directory stack.
This command queries the shell with the command bound to 
`shell-dirstack-query' (default \"dirs\"), reads the next
line output and parses it to form the new directory stack.
DON'T issue this command unless the buffer is at a shell prompt.
Also, note that if some other subprocess decides to do output
immediately after the query, its output will be taken as the
new directory stack -- you lose. If this happens, just do the
command again."
  (interactive)
  (let* ((proc (get-buffer-process (current-buffer)))
	 (pmark (process-mark proc)))
    (goto-char pmark)
    (sit-for 0) ; force redisplay
    (comint-send-string proc shell-dirstack-query) 
    (comint-send-string proc "\n")
    (set-marker pmark (point))
    (let ((pt (point))) ; wait for 1 line
      ;; This extra newline prevents the user's pending input from spoofing us.
      (insert "\n") (backward-char 1)
      (while (not (looking-at ".+\n"))
	(accept-process-output proc)
	(goto-char pt))
      (if (looking-at (concat shell-dirstack-query "\n"))
	  (accept-process-output proc)))
    (goto-char pmark) (delete-char 1) ; remove the extra newline
    (forward-line -1)
    ;; That's the dirlist. grab it & parse it.
;    (let* ((dl (buffer-substring (match-beginning 0) (1- (match-end 0))))
    (let* ((dl (buffer-substring (point) (save-excursion (end-of-line) (point))))
	   (dl-len (length dl))
	   (ds nil)			; new dir stack
	   (i 0)
	   new-dir)
      (while (< i dl-len)
	;; regexp = optional whitespace, (non-whitespace), optional whitespace
	(string-match "\\s *\\(\\S +\\)\\s *" dl i) ; pick off next dir
	(setq new-dir (concat comint-file-name-prefix
			      (substring dl (match-beginning 1) (match-end 1))))
	(setq i (match-end 0))
	(if (not (string-match "/$" new-dir))
	    (setq new-dir (concat new-dir "/")))
	(setq ds (cons new-dir ds)) )
      (goto-char (process-mark proc))
      (let ((ds (nreverse ds)))
	(condition-case nil
	    (progn (shell-cd (car ds))
		   (setq shell-dirstack (cdr ds))
		   (my-shell-dirstack-message t))
	  (error (message (format "Could not cd to %s." (car ds)))))))))


(defun my-shell-directory-tracker (str)
  "Tracks cd, pushd and popd commands issued to the shell.
This function is called on each input passed to the shell.
It watches for cd, pushd and popd commands and sets the buffer's
default directory to track these commands.

You may toggle this tracking on and off with M-x dirtrack-toggle.
If emacs gets confused, you can resync with the shell with M-x dirs.

See variables `shell-cd-regexp', `shell-pushd-regexp', and `shell-popd-regexp',
while `shell-pushd-tohome', `shell-pushd-dextract' and `shell-pushd-dunique'
control the behavior of the relevant command.

Environment variables are expanded, see function `substitute-in-file-name'."
  (if shell-dirtrackp
      ;; We fail gracefully if we think the command will fail in the shell.
      (condition-case chdir-failure
	  (let ((start (progn (string-match "^[; \t]*" str) ; skip whitespace
			      (match-end 0)))
		(def-dir  default-directory)
		end cmd arg1)
	    (while (string-match shell-command-regexp str start)
	      (setq end (match-end 0)
		    cmd (comint-arguments (substring str start end) 0 0)
		    arg1 (comint-arguments (substring str start end) 1 1))
	      (cond ((string-match (concat "\\`\\(" shell-popd-regexp "\\)\\($\\|[ \t]\\)") cmd)
		     (shell-process-popd (substitute-in-file-name arg1)))
		    ((string-match (concat "\\`\\(" shell-pushd-regexp "\\)\\($\\|[ \t]\\)") cmd)
		     (shell-process-pushd (shell-map-home-aliases
					   (shell-replace-stack-dir
					    (substitute-in-file-name arg1))
					   t)))
		    ((string-match (concat "\\`\\(" shell-cd-regexp "\\)\\($\\|[ \t]\\)") cmd)
		     (shell-process-cd (shell-map-home-aliases
					(shell-replace-stack-dir
					 (substitute-in-file-name arg1))
					t))))
	      (if (equal def-dir default-directory)
		  nil
		(setq shell-last-dir def-dir)
		(setq def-dir default-directory))
	      (setq start (progn (string-match "[; \t]*" str end) ; skip again
				 (match-end 0)))))
	    (error (message (format "Could not cd: %s" (car (cdr chdir-failure))))))))

(defun shell-replace-stack-dir (str &optional nomsg defdir)
  "Replace reference to directory stack in given argument by its
value.  Leading parts of STR matching -, =n, =- (for n >= 0) are
replaced by the corresponding directory element (last dir, nth on
stack, last on stack, resp); otherwise STR itself is returned. Strings
of the form +n are not processed since, depending on variables like
'shell-pushd-dextract', these might require special treatment.
Replacement handles the addition of '/', so it is not necessary that
directory elements end in '/'.  If n above is larger than the
directory stack length, a message to that effect is printed, unless
NOMSG is non-nil.  When an ERROR occurs or when a 0 reference into the
stack is given, returns DEFDIR, which defaults to default-directory if
not provided."
  (let* ((def-dir (or defdir default-directory))
	 (stack   (cons def-dir shell-dirstack))
	 (stacklen (length stack))
	 (num (and (stringp str)
		   (or (and (string-match "^=\\([0-9]+\\)\\(/.*\\|\\s \\|$\\)" str)
			    (string-to-int (substring str (match-beginning 1) (match-end 1))))
		       (and (string-match "^=-\\(/.*\\|\\s \\|$\\)" str) (1- stacklen)))))
	 dir-head
	 dir-tail)
    (cond ((and (stringp str) (string-equal "-" str)) shell-last-dir)
	  ((null num) str)
	  ((>= num stacklen)
	   (if (not nomsg)
	       (signal 'error
		       (list (format "Directory stack not that deep (%d)." num))))
	   def-dir)
	  ((string-match "^=\\(\\([0-9]+\\|-\\)/?\\)" str) ;; assumes dirstack entries end in '/'
	   (setq dir-head (nth num stack))
	   (setq dir-tail (substring str (match-end 1)))
	   (concat dir-head (if (not (string-match "/$" dir-head)) "/") dir-tail))
	  (t str))))


;;; cd [dir]
(defun my-shell-process-cd (arg)
  (let ((new-dir (cond ((zerop (length arg)) (concat comint-file-name-prefix
						     "~/"))
		       ((string-equal "-" arg) shell-last-dir)
		       (t (shell-prefixed-directory-name arg)))))
    (shell-cd new-dir)
    (shell-dirstack-message)))

;;; pushd [+n | dir]
(defun my-shell-process-pushd (arg)
  (let ((num (shell-extract-num arg)))
    (cond ((zerop (length arg))
	   ;; no arg -- swap pwd and car of stack unless shell-pushd-tohome
	   (cond (shell-pushd-tohome
		  (shell-process-pushd (concat comint-file-name-prefix "~/")))
		 (shell-dirstack
		  (let ((old default-directory))
		    (shell-cd (car shell-dirstack))
		    (setq shell-dirstack
			  (cons old (cdr shell-dirstack)))
		    (shell-dirstack-message)))
		 (t
		  (message "Directory stack empty."))))
	  ((numberp num)
	   ;; pushd +n
	   (cond ((> num (length shell-dirstack))
		  (message "Directory stack not that deep."))
		 ((= num 0)
		  (error (message "Couldn't cd.")))
		 (shell-pushd-dextract
		  (let ((dir (nth (1- num) shell-dirstack)))
		    (shell-process-popd arg)
		    (shell-process-pushd dir)))
		 (t
		  (let* ((ds (cons default-directory shell-dirstack))
			 (dslen (length ds))
			 (front (nthcdr num ds))
			 (back (reverse (nthcdr (- dslen num) (reverse ds))))
			 (new-ds (append front back))
			 (new-dir (car new-ds)))
		    (shell-cd new-dir)
		    (setq shell-dirstack (cdr new-ds))
		    (shell-dirstack-message)))))
	  (t
	   ;; pushd <dir>
	   (let* ((old-wd default-directory)
		 (dir (shell-prefixed-directory-name arg))
		 (ex-dir (expand-directory-name dir)))
	     (if (not (string-match "/$" dir))
		 (setq dir (concat dir "/")))
	     (shell-cd dir)
	     (setq shell-dirstack (cons old-wd shell-dirstack))
	     (if (and shell-pushd-dunique
		      (member ex-dir
			      (mapcar 'expand-directory-name shell-dirstack)))
		 (setq shell-dirstack
		       (remove-matching-elements ex-dir shell-dirstack
						 'equal-dir-name))))
	   (shell-dirstack-message)))))


(defun my-shell-dirstack-message (&optional show)
  "Show the current dirstack on the message line.
If the variable shell-pushd-silent is non-nil, then no message
is shown.  This mimicks the optional behavior of tcsh."
  (if (and shell-pushd-silent (null show))
      nil
    (let* ((msg "")
	   (ds (cons default-directory shell-dirstack))
	   (home (expand-file-name (concat comint-file-name-prefix "~/")))
	   (homelen (length home)))
      (while ds
	(let ((dir (car ds)))
	  (and (>= (length dir) homelen) (string= home (substring dir 0 homelen))
	       (setq dir (concat "~/" (substring dir homelen))))
	  ;; Strip off comint-file-name-prefix if present.
	  (and comint-file-name-prefix
	       (>= (length dir) (length comint-file-name-prefix))
	       (string= comint-file-name-prefix
			(substring dir 0 (length comint-file-name-prefix)))
	       (setq dir (substring dir (length comint-file-name-prefix)))
	       (setcar ds dir))
	  (setq msg (concat msg (directory-file-name dir) " "))
	  (setq ds (cdr ds))))
      (message "%s" msg))))

(defun my-shell-replace-by-expanded-directory ()
  "Expand directory stack reference before point.
Directory stack references are of the form \"=digit\" or \"=-\".
See `default-directory' and `shell-dirstack'.

Now handles trailing / correctly.

Returns t if successful."
  (interactive)
  (if (comint-match-partial-filename)
      (save-excursion
	(goto-char (match-beginning 0))
	(let ((stack (cons default-directory shell-dirstack))
	      (index (cond ((looking-at "=-/?")
			    (length shell-dirstack))
			   ((looking-at "=\\([0-9]+\\)/?")
			    (string-to-number
			     (buffer-substring
			      (match-beginning 1) (match-end 1)))))))
	  (cond ((null index)
		 nil)
		((>= index (length stack))
		 (error "Directory stack not that deep."))
		(t
		 (replace-match (file-name-as-directory (nth index stack)) t t)
		 (message "Directory item: %d" index)
		 t))))))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                                                           ;;
;;(@* "File and Navigation Modes")                           ;;;;;;
;;                                                           ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Dired-mode

(setq-default dired-listing-switches "-lF")

(defun my-dired-load-hook ()
  (defun dired-find-file-read-only ()
    "In dired, visit the file or directory named on this line in read-only state."
    (interactive)
    (let ((file-name (file-name-sans-versions (dired-get-filename) t)))
      (if (file-exists-p file-name)
          (find-file-read-only file-name)
        (if (file-symlink-p file-name)
            (error "File is a symlink to a nonexistent target")
          (error "File no longer exists; type `g' to update Dired buffer"))))
    )
  (defun dired-mac-open ()
    "In dired, run open command on given file or on each file in list"
    (interactive)
    (dired-do-shell-command "open")
    )
  (defalias 'dired-advertised-find-file-read-only 'dired-find-file-read-only)
  (setq dired-ls-F-marks-symlinks t) ;; Says ls -F marks symlinks with @
  (setq dired-listing-switches "-lF")
  )

(defun my-dired-mode-hook ()
  (local-set-key "\C-m"  'my-dired-find-file-or-subdir)
  (local-set-key "\M-c"  'my-kill-ring-save)
  (local-set-key [?\M-C] 'diredp-capitalize-this-file)
  (local-set-key "r"     'dired-advertised-find-file-read-only)
  (local-set-key "F"     'find-dired)
  (local-set-key "G"     'find-grep-dired)
  (local-set-key "N"     'find-name-dired)
  (local-set-key "%F"    'find-dired)
  (local-set-key "%G"    'find-grep-dired)
  (local-set-key "%N"    'find-name-dired)
  (local-set-key "W"     'browse-url-of-dired-file)
  (local-set-key "\M-o"  'dired-mac-open)
  (setq dired-listing-switches "-lF")
  )

(add-hook 'dired-load-hook 'my-dired-load-hook)
(add-hook 'dired-mode-hook 'my-dired-mode-hook)

(if (not (featurep 'dired+))
    (message "Dired+ is not loaded")
  (set-face-background 'diredp-no-priv     "white")
  (set-face-background 'diredp-write-priv  "tan")
  (set-face-background 'diredp-dir-heading "bisque1")
  (set-face-background 'diredp-dir-priv    "white")
  (set-face-foreground 'diredp-dir-priv    "#FF3445")
  (set-face-foreground 'diredp-file-name   "#3D34FF")
  (set-face-foreground 'diredp-file-suffix "#CF34FF")
  )


(defun my-dired-find-file-or-subdir ()
  "In dired, visit the file or sub-directory named on this line.
If the name corresponds to a sub-directory we use dired-maybe-insert-subdir;
otherwise, we use dired-find-file.  This allows directory navigation to all
occur within the same buffer. \\[dired-find-file] still works as before
if separate buffers are desired."
  (interactive)
  (let ((dir  (file-name-as-directory (dired-get-filename))))
    (if (not (file-directory-p dir))
        (dired-find-file)
      (dired-maybe-insert-subdir dir
                                 (if current-prefix-arg
                                     (read-string "Switches for listing: " dired-actual-switches)))
      (dired-goto-next-file))
    ))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Linkd Mode

(require-soft 'linkd)
(setq linkd-use-icons t)
(setq linkd-icons-directory (concat my-site-lisp-dir "other-modes/linkd-icons"))
(setq linkd-wiki-directory "~/.emacs.d/linkd-wiki")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Icicles

(require-soft 'icicles)
(when (featurep 'icicles)
  (add-to-list 'icicle-modal-cycle-up-keys              "\C-p")
  (add-to-list 'icicle-modal-cycle-up-action-keys       "\C-\M-p")
  (add-to-list 'icicle-modal-cycle-up-alt-action-keys   [?\C-\S-p])
  (add-to-list 'icicle-modal-cycle-up-help-keys         [?\C-\M-\S-p])
  (add-to-list 'icicle-modal-cycle-down-keys            "\C-n")
  (add-to-list 'icicle-modal-cycle-down-action-keys     "\C-\M-n")
  (add-to-list 'icicle-modal-cycle-down-alt-action-keys [?\C-\S-n])
  (add-to-list 'icicle-modal-cycle-down-help-keys       [?\C-\M-\S-n])
  )

(defvar my-icicle-describe-sort-in-completions (and t (featurep 'icicles))
  "If non-nil, show the current sort order in *Completions* buffer.")

(defun my-icicle-show-sort-order ()
  "Show current sort order in completions list.
   Add this to completion-setup-hook to achieve
   the effect."
  (when (and my-icicle-describe-sort-in-completions icicle-mode)
    (with-current-buffer standard-output
      (let* ((sort-order (icicle-current-sort-order nil))
             (sort-string (concat ", sorting "
                                  (if (string-equal sort-order "alphabetical")
                                      "alphabetically" sort-order))))
        (goto-char (point-min))
        (when (re-search-forward "Possible completions are" nil t)
          (put-text-property 0 (length sort-string) 'face 'icicle-Completions-instruction-1 sort-string)
          (insert sort-string)))))
  )

(defun icicle-beginning-of-Completions ()
  "Scroll to the beginning of *Completions* window."
  (interactive)
  (let ((window (get-buffer-window "*Completions*" 0)))
  (when window
    (save-selected-window
      (select-window window)
      (unless (= (window-start) (point-min))
        (goto-char (icicle-start-of-candidates-in-Completions)))
      ))))

(defun icicle-end-of-Completions ()
  "Scroll to the end of the *Completions* window, showing
   as many of the completions as possible."
  (interactive)
  (let ((window (get-buffer-window "*Completions*" 0)))
  (when window
    (save-selected-window
      (select-window window)
      (unless (= (window-end) (point-max))
        (goto-char (point-max))
        (scroll-down (1- (/ (window-height) 2)))
        (beginning-of-line))
      ))))

(defun my-icicle-mode-hook ()
  (setq read-file-name-completion-ignore-case t) ; Emacs >= 22
  (setq read-buffer-completion-ignore-case t)    ; Emacs >= 23
  (setq icicle-show-Completions-help-flag nil)
  (when (and icicle-mode (lookup-key icicle-mode-map "\C-h"))
    (map-keymap
     (lambda (event binding)
       (define-key icicle-mode-map
         (vector ?\C-c ?\C-h event) binding))
     (lookup-key icicle-mode-map "\C-h"))
    (define-key icicle-mode-map "\C-h" nil))
  (if icicle-mode
      (add-hook 'completion-setup-hook 'my-icicle-show-sort-order t)
    (remove-hook 'completion-setup-hook 'my-icicle-show-sort-order))
  )

(add-hook 'icicle-mode-hook 'my-icicle-mode-hook)

(defun my-icicle-minibuffer-setup-hook ()
  (define-key minibuffer-local-completion-map "\C-w" 'icicle-scroll-Completions-up)
  (define-key minibuffer-local-completion-map "\C-v" 'icicle-scroll-Completions)
  (define-key minibuffer-local-completion-map "\M-w" 'icicle-beginning-of-Completions)
  (define-key minibuffer-local-completion-map "\M-v" 'icicle-end-of-Completions)
  (define-key minibuffer-local-completion-map "\M-k" 'icicle-kill-region)
  (define-key minibuffer-local-completion-map "\M-e" 'icicle-erase-minibuffer-or-history-element)
  (define-key minibuffer-local-completion-map "\C-x\C-o" 'icicle-switch-to/from-minibuffer)
  ;; process new help key position
  (define-key minibuffer-local-completion-map "\M-y" 'icicle-history)
  (define-key minibuffer-local-completion-map "\M-h" nil)
  ;; the following no longer appear necessary
  ;(define-key minibuffer-local-completion-map "\C-n" 'icicle-next-candidate-per-mode)
  ;(define-key minibuffer-local-completion-map "\C-p" 'icicle-previous-candidate-per-mode)
  )
(add-hook 'icicle-minibuffer-setup-hook 'my-icicle-minibuffer-setup-hook)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Interactive Do Things (ido-mode)

(require-soft 'ido)
(defvar my-ido-begin-str "~/")
(defvar my-ido-begin-home 'home
  "*Determines the behavior of \\[my-ido-begin-line].
    If equal 'home,  replaces existing
    file text with the home directory. If 'root,
    replaces it with root. And if 'old, replaces
    it with value of \\[my-ido-begin-str]. The
    function toggles among these states.")

(defun my-ido-begin-line ()
  "*Moves to the beginning of the find file line,
    toggling between the home directory and root
    on successive calls."
  (interactive)
  (let
      ;; ATTN: Fix this 12 == length("Find file: ")
      ((contents (buffer-substring-no-properties 12 (minibuffer-prompt-end))))
    (cond
     ((equal my-ido-begin-home 'home)
      (insert "~/")
      (setq my-ido-begin-home 'root)
      ;;(setq my-ido-begin-str contents)
      (kill-new contents) ;;ATTN: for now make old string available to yank
      )
     ((equal my-ido-begin-home 'root)
      (insert "//")
      (setq my-ido-begin-home 'home) ;;'old)
      )
     (t ;; ATTN: for now this isn't working because delete fails so ignore it
      (setq my-ido-begin-home 'home)
      ;;(delete-region 12 (minibuffer-prompt-end))
      ;;(insert my-ido-begin-str)
      )))
  (setq ido-rescan t)
  )

(defun my-ido-setup-hook ()
  "Adjust keymaps for ido completion and other customizations."
  (define-key ido-file-completion-map "\C-a" 'my-ido-begin-line)
  (define-key ido-file-completion-map "\C-p" 'ido-prev-match)
  (define-key ido-file-completion-map "\C-n" 'ido-next-match)
  (define-key ido-file-completion-map "\C-xc"    'ido-toggle-case)
  (define-key ido-file-completion-map "\C-x\C-c" 'ido-toggle-case)
  (define-key ido-file-completion-map "\C-x\C-i" 'ido-toggle-ignore)
  (define-key ido-file-completion-map "\C-xi"    'ido-toggle-ignore)
  (define-key ido-file-completion-map "\C-xp"    'ido-toggle-prefix)
  (define-key ido-file-completion-map "\C-x\C-p" 'ido-toggle-prefix)
  (define-key ido-file-completion-map "\C-xr"    'ido-toggle-regexp)
  (define-key ido-file-completion-map "\C-x\C-r" 'ido-toggle-regexp)
  (setq ido-enable-flex-matching t)
  )

(add-hook 'ido-setup-hook 'my-ido-setup-hook)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Anything


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; YASnippet

(require-soft 'yasnippet)

(when (featurep 'yasnippet)
  (setq yas/root-directory "~/.emacs.d/snippets")
  (yas/initialize)
  (yas/load-directory yas/root-directory)
  (setq yas/prompt-functions '(yas/ido-prompt yas/dropdown-prompt yas/completing-prompt yas/x-prompt yas/no-prompt)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Which-Function

(require-soft 'which-func)
(setq which-func-modes 
      '(emacs-lisp-mode c-mode c++-mode python-mode perl-mode cperl-mode makefile-mode sh-mode fortran-mode f90-mode ada-mode))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; View-mode

(defun my-view-mode-hook ()
  (define-key view-mode-map "\M-v" 'end-of-buffer)
  (define-key view-mode-map "k"    'View-kill-and-leave)
  (define-key view-mode-map "l"    'View-leave)
  )

(add-hook 'view-mode-hook 'my-view-mode-hook)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Help-mode

(defun my-help-mode-hook ()
  (local-set-key ","    'rename-uniquely)
  (local-set-key "a"    'command-apropos)
  (local-set-key "\M-a" 'apropos-documentation)
  (local-set-key "b"    'describe-bindings)
  (local-set-key "c"    'describe-key-briefly)
  (local-set-key "C"    'describe-coding-system)
  (local-set-key "d"    'apropos-documentation)
  (local-set-key "e"    'view-echo-area-messages)
  (local-set-key "f"    'describe-function)
  (local-set-key "F"    'Info-goto-emacs-command-node)
  (local-set-key "i"    'info)
  (local-set-key "k"    'describe-key)
  (local-set-key "K"    'Info-goto-emacs-key-command-node)
  (local-set-key "l"    'help-go-back) ; I just always want to do this, so...
  (local-set-key "L"    'view-lossage) ; minor inconsistency won't hurto
  (local-set-key "\M-l" 'locate-library)
  (local-set-key "m"    'describe-mode)
  (local-set-key "p"    'finder-by-keyword)
  (local-set-key "q"    'View-leave)
  (local-set-key "\M-q" 'View-kill-and-leave)
  (local-set-key "r"    'info-emacs-manual)
  (local-set-key "s"    'describe-syntax)
  (local-set-key "S"    'info-lookup-symbol)
  (local-set-key "u"    'manual-entry)
  (local-set-key "v"    'describe-variable)
  (local-set-key "V"    'describe-language-environment) ; another minor inconsistency
  (local-set-key "w"    'where-is)
  (local-set-key [\M-left] 'help-go-back)
  )

(add-hook 'help-mode-hook 'my-help-mode-hook)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Doc-View Mode

(setq doc-view-ghostscript-program "/opt/local/bin/gs")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Hippie Expand

(condition-case nil
    (delete 'try-expand-line hippie-expand-try-functions-list)
   ;(delete 'try-expand-list hippie-expand-try-functions-list)
   (error nil))
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                                                           ;;
;;(@* "TeX-related Modes")                                   ;;
;;                                                           ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; AUC TeX support

(setq tex-default-mode 'plain-tex-mode) ;includes non AucTeX case
(setq TeX-default-mode 'plain-TeX-mode)
(setq my-TeX-generic-setup-done nil)

(defun my-generic-TeX-setup ()
  "Configure global settings and interface that affects all 
AucTeX modes. Only performed once across all modes."
  (do-only-once my-TeX-generic-setup-done
    (add-to-list 'TeX-format-list '("TEX" plain-tex-mode "\\input\\b"))
    (add-to-list 'TeX-expand-list '("%a" file "pdf" t))
    (add-to-list 'TeX-expand-list '("%M"
                                    (lambda ()
                                      (let ((targets (my-get-makefile-targets))) 
                                        (completing-read "Target: " targets)) )))
    (delete-if (lambda (item) (string-equal "Makeinfo" (car item))) TeX-command-list)
    (delete-if (lambda (item) (string-equal "Makeinfo HTML" (car item))) TeX-command-list)
    (add-to-list 'TeX-command-list '("Info" "makeinfo %t" TeX-run-compile nil
                                     (texinfo-mode) :help "Run makeinfo with Info output"))
    (add-to-list 'TeX-command-list '("Info HTML" "makeinfo --html %t" TeX-run-compile nil
                                     (texinfo-mode) :help "Run makeinfo with HTML output"))
    (add-to-list 'TeX-command-list '("PDF" "%(o?)dvipdf %d %a " TeX-run-command t t
                                     :help "Generate PDF file"))
    (add-to-list 'TeX-command-list '("Make" "make %M " 'TeX-run-command t t
                                     :help "Generate with a Makefile"))
    (require 'texmathp)
    (eval-after-load 'texmathp
      (progn
        (setq texmathp-tex-commands '(("equations" env-on)
                                      ("\\halign"  arg-off)))
        (texmathp-compile) ))
    (setq TeX-parse-self t)
    (setq TeX-auto-save  t)))

(defun my-plain-TeX-mode-hook ()
  "Personal configuration for GenTeX and plain TeX."
  (local-set-key "\C-cp" 'TeX-PDF-mode)
  (local-set-key "\C-c[" 'TeX-insert-braces) ; why shift?
  (local-set-key "\C-c\C-n" nil)
  (local-set-key "\C-c\C-m" nil)
  (local-set-key [?\C-c return] 'TeX-insert-macro)
  ;; electric $ balks on embedded $ but try again for now
  ;(local-set-key "$" 'self-insert-command)
  ;; GenTeX specific font specifiers
  (setq TeX-font-list '( (?\C-b "\\textbold{" "}" "\\mathbold{" "}") 
			 (?\C-c "\\smcaps{" "}")
			 (?\C-e "\\emph{" "}")
			 (?\C-i "{\\it " "\\/}")
			 (?\C-r "{\\rm " "}")
			 (?\C-s "\\sans{" "}")
			 (?\C-t "\\tty{" "}")
			 (?\C-u "{" "}")
			 (?\C-d "" "" t)
			 (?\M-p "{\\xlarge " "}")
			 (?\C-p "{\\large " "}")
			 (?\C-m "{\\medium " "}")
			 (?\C-n "{\\small " "}")
			 (?\M-n "{\\xsmall " "}")
			 ))
  (LaTeX-add-environments '("equations" LaTeX-env-label)) 
  (local-set-key "\C-c\C-e" 'LaTeX-environment)
  (setq LaTeX-default-environment "itemize"))
  
(add-hook 'plain-TeX-mode-hook 'highlight-attn-words)
(add-hook 'plain-TeX-mode-hook 'turn-on-flyspell) 
(add-hook 'plain-TeX-mode-hook 'turn-on-tex-parser-ispell)
(add-hook 'plain-TeX-mode-hook 'my-plain-TeX-mode-hook)
(add-hook 'plain-TeX-mode-hook 'my-generic-TeX-setup)

(add-hook 'LaTeX-mode-hook 'highlight-attn-words)
(add-hook 'LaTeX-mode-hook 'turn-on-tex-parser-ispell)
(add-hook 'LaTeX-mode-hook 'turn-on-flyspell) 
(add-hook 'LaTeX-mode-hook 'my-generic-TeX-setup)

(autoload 'LaTeX-environment "latex" "Make Environment Using AucTeX commands" t)
(autoload 'LaTeX-add-environments "latex" "Add Environment Info Using AucTeX commands" t)

(add-hook 'AmS-TeX-mode-hook 'my-generic-TeX-setup)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; BibTeX-mode

(setq bibtex-mode-hook
      '((lambda ()
	  (local-set-key "\"" 'self-insert-command)
	  (setq bibtex-include-OPTannote nil)
	  (setq bibtex-include-OPTcrossref nil)
	  )))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                                                           ;;
;;(@* "Other Modes and Features")                            ;;;;;;
;;                                                           ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Formats

(defvar my-date-format "%Y %b %d %a"
  "My favorite date format; see format-time-string for details.")
(setq header-date-format "%Y %b %d %a %H:%M:%S")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; IBuffer

(defun ibuffer-list-buffers-select (&optional noselect)
  "Display a list of buffers, in another window.
If optional argument FILES-ONLY is non-nil, then add a filter for
buffers which are visiting a file."
  (interactive "P")
  (let ((starting-buffer (buffer-name (current-buffer))))
    (ibuffer t nil nil noselect)
    (if (and (null noselect)
             ibuffer-start-at-most-recent
             (eq major-mode 'ibuffer-mode))  ; only do this in the ibuffer window
        (ibuffer-jump-to-buffer starting-buffer))))

(defun ibuffer-match-my-files (label &optional subdir pattern)
  "Create ibuffer filter groups associated with specified file or
subdirectory names in my home directory. Files in the group are
those that match the path, which can be only partially specified.
LABEL is the ibuffer name associated with the filter group.
SUBDIR is a subdirectory path that restricts the files selected
if non-nil. It should not start with / in general.
PATTERN is a regexp that limits the files considered within
the subdirectory.
   The return value of the function is a list containing the
list entry specifying the filter group, so that it can be included
in the filter group list with ,@ and will not affect the resulting
list if no such files exist."
  (let* ((subdir (or subdir ""))
         (nodes  (mapcar (lambda (x) `(filename . ,x))
                         (directory-files (concat my-home-dir subdir) t pattern))))
    (if (= (length nodes) 0) nil
      (list (cons label (if (= (length nodes) 1) nodes (list (cons 'or nodes))))))))

(defvar ibuffer-start-at-most-recent nil
  "If non-nil, \\[ibuffer-list-buffers-select] will, if selecting the
ibuffer listing window, put the cursor on the line corresponding to the
most recent buffer. Otherwise, the cursor is positioned on the top
buffer line in the ibuffer listing.")

(defvar my-ibuffer-loaded nil
  "Variable used to control loading of ibuffer customized setup.
Should always be defined as NIL initially.")

(defun my-ibuffer-mode-hook () 
  ;; Setup definitions and filter groups.
  ;; To avoid worrying about requires, handle the setup
  ;; in the hook but only on the first time called.
  (do-only-once my-ibuffer-loaded
    (require 'ibuf-ext) ;; Only required as need and when ibuffer exists
    (define-ibuffer-filter mode-nostar
      "Toggle current view to buffers with major mode QUALIFIER
     and no star at beginning of the buffer name."
      (:description "major mode with no star in name"
                    :reader
                    (intern
                     (completing-read "Filter by major mode: " obarray
                                      #'(lambda (e)
                                          (string-match "-mode$"
                                                        (symbol-name e)))
                                      t
                                      (let ((buf (ibuffer-current-buffer)))
                                        (if (and buf (buffer-live-p buf))
                                            (with-current-buffer buf
                                              (symbol-name major-mode))
                                          "")))))
      (and
       (not (string-match "^\\s-*\\*" (buffer-name buf)))
       (eq qualifier (with-current-buffer buf major-mode))))
    (setq ibuffer-saved-filter-groups
          `(("default"
             ,@(ibuffer-match-my-files "Teaching" "class/" "^s")
             ("Documents" (or (mode . plain-tex-mode)
                              (mode . latex-mode)
                              (mode . bibtex-mode)
                              (mode . amstex-mode)
                              (mode-nostar . org-mode)
                              (mode-nostar . fundamental-mode)
                              (mode . nxml-mode)
                              (mode . nxhtml-mode)))
             ,@(ibuffer-match-my-files "Android" "Documents/" "^Eclipse")
             ,@(ibuffer-match-my-files "Arduino" "Programming/" "^Arduino")
             ("Code" (or (mode . cc-mode)
                         (mode . python-mode)
                         (mode . java-mode)
                         (mode . ess-mode)
                         (mode . js2-mode)
                         (mode . ruby-mode)
                         (mode . haskell-mode)
                         (mode . perl-mode)
                         (mode . tuareg-mode)
                         (mode . R-mode)
                         (mode . r-mode)
                         (mode . R-transcript-mode)
                         (mode . r-transcript-mode)
                         (filename . "[Mm]akefile")))
             ("Emacs" (mode-nostar . emacs-lisp-mode))
             ("Shells" (or (mode . shell-mode)
                           (mode . inferior-ess-mode)
                           (name . "^\\*[iI][pP]ython\\*")
                           (mode . lisp-interaction-mode)))
             ("Dirs"  (mode . dired-mode))
             ("Help"  (or (name . "^\\*Help\\*")
                          (name . "^\\*help\\[R\\]")
                          (name . "^\\*Man\\s-")
                          (name . "^\\*Apropos\\*")
                          (mode . help-mode)
                          (mode . Man-mode)
                          (mode . Info-mode)))
             ("Utility" (name . "^\\*.*\\*\\(<[0-9]+>\\)?$"))
             ("Other" (name . "^\\S-")))
                  ("all"
                   ("All" (name . ".")))
                  )))
  ;; Normal operational settings.
  (setq ibuffer-start-at-most-recent t)
  (ibuffer-auto-mode 1)
  (local-set-key "\M-\C-f" 'ibuffer-forward-filter-group)
  (local-set-key "\M-\C-b" 'ibuffer-backward-filter-group)
  (local-set-key "/F"      (lambda () (interactive) (ibuffer-filter-by-filename ".*")))
  (setq ibuffer-show-empty-filter-groups nil)
  (ibuffer-switch-to-saved-filter-groups "default"))

(add-hook 'ibuffer-mode-hook 'my-ibuffer-mode-hook)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Intelligent Copy and Kill

  ;; ATTN: not done yet

(defvar my-noregion-copykill-thing-default-order
  '((url . nearest)
    (word . nearest)
    (line . at)
    (page . at)
    (sentence . at)
    (paragraph . at)
    (sexp . nearest)
    (defun . nearest))
  "A list defining the order and scope of things
that are captured by smart copy and kill operations.
Each element of the list is a cons cell of the form
(THING . SCOPE) where THING is a symbol that corresponds
to a capturable thing in thingatpt/thingatpoint+ and
SCOPE is either the symbol at or nearest, corresponding
to the functions thing-at-point or thing-nearest-point.
The nearest scope only works if thingatpt+ is available,
otherwise it is transformed to at. The default")

 
(defvar my-noregion-copykill-thing-order-alist
  '((emacs-lisp-mode .
      ((sexp . nearest)
       (defun . nearest)
       (line . at)
       (page . at)
       (word . nearest)
       (sentence . at))))
  "")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                                                           ;;
;;(@* "ESS Mode for R and S Code")                           ;;;;;;
;;                                                           ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; ESS-mode for S 

;;
;; Formatting Style 
;;

(setq ESS-crg-style '(CRG (ess-indent-level . 4)
                          (ess-continued-statement-offset . 4)
                          (ess-continued-brace-offset . -4 )
                          (ess-brace-offset . 0)
                          (ess-brace-imaginary-offset . 0)
                          (ess-close-brace-offset . 0)
                          (ess-else-offset . 0)
                          (ess-expression-offset . 0)
                          (ess-arg-function-offset . 0)
                          (ess-fancy-comments . nil)
                          (ess-auto-newline . t)
                          (ess-tab-always-indent . t)
                   ))

;;
;; Basic Configuration
;;

(defun my-ess-mode-load-hook ()
  (if (not (assoc 'CRG ess-style-alist))
      (setq ess-style-alist (cons ESS-crg-style ess-style-alist)))
  (setq ess-default-style 'CRG)
  (setq ess-style 'CRG)
  (setq ess-auto-newline t)
  (setq-default ess-loop-timeout '1000000)
  (setq-default inferior-ess-help-command "help(\"%s\")\n")
  (setq ess-use-toolbar nil)
  (setq ess-fancy-comments nil)
  )

(defun my-inferior-ess-mode-hook ()
  (local-set-key "\C-d" 'delete-char)
  (local-set-key "\C-c\M-p" 'comint-backward-matching-input)
  (local-set-key "\C-c\M-n" 'comint-forward-matching-input)
  (add-hook 'comint-input-filter-functions 'ess-search-path-tracker nil t)
  (setq input-ring-size '1024)
  (setq ess-style 'CRG) ;; these three lines needed?
  (setq ess-auto-newline t)
  (ess-set-style)
  (setq inferior-ess-help-command "help(\"%s\")\n")
  )

(defun my-ess-pre-run-hook ()
  (progn
    (remove-hook 'ess-mode-hook             'turn-on-font-lock)
    (remove-hook 'ess-transcript-mode-hoonk 'turn-on-font-lock)
    (remove-hook 'inferior-ess-mode-hook    'turn-on-font-lock)
    )
  )

(defun my-ess-mode-hook ()
  (setq ess-style 'CRG)
  (setq ess-auto-newline t)
  (ess-set-style)
  )

(defun ess-transcript-mode-hook ()
  (setq ess-style 'CRG)
  (setq ess-auto-newline t)
  (ess-set-style)
  )

;;
;; Load the functions
;;

(add-hook 'ess-mode-load-hook       'my-ess-mode-load-hook)
(add-hook 'inferior-ess-mode-hook   'my-inferior-ess-mode-hook)
(add-hook 'ess-pre-run-hook         'my-ess-pre-run-hook)
(add-hook 'ess-mode-hook            'my-ess-mode-hook)
(add-hook 'ess-transcript-mode-hook 'my-ess-transcript-mode-hook)
(require-soft 'ess-site)

(if (not (featurep 'ess))
    (message "Cannot load ESS."))


;;
;; Disable Font-Lock support in ess-mode
;;

;;;;;Might want to try it for a while 2 Sep 2010
;;;(require 'ess)
;;;
;;;(progn
;;;  (remove-hook 'ess-mode-hook             'turn-on-font-lock)
;;;  (remove-hook 'ess-transcript-mode-hoonk 'turn-on-font-lock)
;;;  (remove-hook 'inferior-ess-mode-hook    'turn-on-font-lock)
;;;  )
;;;
;;;(setq inferior-ess-font-lock-input nil)  ;; needed?

;; (setq ess-style-alist (cons ESS-crg-style ess-style-alist)) ;;  needed?



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                                                           ;;
;;(@* "Global Key Bindings")                                 ;;;;;;
;;                                                           ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;
;; Auxilliary Global Keymaps
;;
    ;; String Replacement Commands

(define-prefix-command 'my-string-replace-prefix 'my-string-replace-map "String Replacement")
(define-key my-string-replace-map "q" 'query-replace)
(define-key my-string-replace-map "x" 'query-replace-regexp)
(define-key my-string-replace-map "s" 'replace-string)
(define-key my-string-replace-map "r" 'replace-regexp)
(define-key my-string-replace-map "t" 'tags-query-replace)

(when (featurep 'icicles)
  (define-key my-string-replace-map [escape] (make-sparse-keymap)) ; to allow icicle-complete-keys
  (add-to-list 'icicle-keymaps-for-key-completion 'my-string-replace-map t))

    ;; Transpose Commands

(define-prefix-command 'my-transpose-prefix 'my-transpose-map "Transpose Commands")
(define-key my-transpose-map "c"   'transpose-chars)
(define-key my-transpose-map "l"   'transpose-lines)
(define-key my-transpose-map "p"   'transpose-paragraphs)
(define-key my-transpose-map "s"   'transpose-sentences)
(define-key my-transpose-map "w"   'transpose-words)
(define-key my-transpose-map "x"   'transpose-sexps)

    ;; Tag Management Commands

(define-prefix-command 'my-tags-prefix 'my-tags-map "Tag Management")
(define-key my-tags-map "l"   'list-tags)
(define-key my-tags-map "."   'find-tag)
(define-key my-tags-map "f"   'find-tag)
(define-key my-tags-map "r"   'find-tag-regexp)
(define-key my-tags-map "4"   'find-tag-other-window)
(define-key my-tags-map "5"   'find-tag-other-frame)
(define-key my-tags-map "n"   'find-tag-noselect)
(define-key my-tags-map "c"   'tags-loop-continue)
(define-key my-tags-map "s"   'tags-search)
(define-key my-tags-map "\C-i" 'complete-tag)
(define-key my-tags-map "t"   'select-tags-table)
(define-key my-tags-map "v"   'visit-tags-table)
(define-key my-tags-map "z"   'tags-reset-tags-table)
(define-key my-tags-map "a"   'tags-apropos)
(define-key my-tags-map "q"   'tags-query-replace)

(when (featurep 'icicles)
  (define-key my-tags-map [escape] (make-sparse-keymap)) ; to allow icicle-complete-keys
  (add-to-list 'icicle-keymaps-for-key-completion 'my-tags-map t))

    ;; Marks and so forth (e.g., Bookmarks)

(defvar my-mark-map (copy-keymap bookmark-map) "Mark/Bookmark Management")
(fset 'my-mark-prefix my-mark-map)

(when (featurep 'icicles)
  (define-key my-mark-map [escape] (make-sparse-keymap)) ; to allow icicle-complete-keys
  (add-to-list 'icicle-keymaps-for-key-completion 'my-mark-map t))

    ;; Go Map  (Goto's and Jumps of All Kinds)

(define-prefix-command 'my-go-prefix 'my-go-map "Go")
(define-key my-go-map "g"      'goto-line)
(define-key my-go-map "\M-g"   'goto-line)
(define-key my-go-map "h"      'help-command)
(when (featurep 'linkd)
  (define-key my-go-map "l"    'linkd-follow-at-point))
(define-key my-go-map "n"      'next-error)
(define-key my-go-map "\M-n"   'next-error)
(when (featurep 'org)
  (define-key my-go-map "o"    'org-open-at-point))
(define-key my-go-map "p"      'previous-error)
(define-key my-go-map "\M-p"   'previous-error)
(define-key my-go-map "q"      'keyboard-quit)  
(define-key my-go-map "t"      'top-level)
(define-key my-go-map "u"      'browse-url)
(define-key my-go-map "x"      'exit-recursive-edit)
(define-key my-go-map [escape] 'keyboard-escape-quit)

(when (featurep 'icicles)
  (define-key my-go-map [escape] (make-sparse-keymap)) ; to allow icicle-complete-keys
  (add-to-list 'icicle-keymaps-for-key-completion 'my-go-map t))

(when (featurep 'semantic)
  (define-prefix-command 'my-go-semantic-prefix 'my-go-semantic-map "Go Places with Semantic")
  (when (featurep 'senator)
    (define-key my-go-semantic-map "j" 'senator-jump)
    (define-key my-go-semantic-map "n" 'senator-next-tag)
    (define-key my-go-semantic-map "p" 'senator-previous-tag)
    (define-key my-go-semantic-map "u" 'senator-go-to-up-reference))
  (define-key my-go-semantic-map "B" 'semantic-mrub-switch-tags)
  (define-key my-go-semantic-map "c" 'semantic-complete-jump)
  (define-key my-go-semantic-map "f" 'semantic-ia-fast-jump)
  (define-key my-go-semantic-map "h" 'semantic-analyze-proto-impl-toggle)
  (define-key my-go-semantic-map "r" 'semantic-symref)
  (define-key my-go-semantic-map "s" 'semantic-symref-symbol)
  (define-key my-go-semantic-map "x" 'semantic-symref-regexp)
  (define-key my-go-map "s" 'my-go-semantic-prefix))

(if (featurep 'bookmark+)
    (progn
      (define-key my-go-map "b"    'bmkp-jump-map)
      (define-key my-go-map "B"    'bmkp-jump-other-window-map))
  (define-key my-go-map "b"    'bookmark-jump)
  (define-key my-go-map "B"    'bookmark-jump-other-window))


    ;; Abbrevs

(define-prefix-command 'my-abbrev-prefix 'my-abbrev-map "Abbreviations")
(define-key my-abbrev-map "c"      'dabbrev-completion)
(define-key my-abbrev-map "d"      'define-global-abbrev)
(define-key my-abbrev-map "\C-d"   'define-mode-abbrev)
(define-key my-abbrev-map "e"      'expand-abbrev)
(define-key my-abbrev-map "g"      'add-global-abbrev)
(define-key my-abbrev-map "\C-g"   'inverse-add-global-abbrev)
(define-key my-abbrev-map "k"      'kill-all-abbrevs)
(define-key my-abbrev-map "l"      'list-abbrevs)
(define-key my-abbrev-map "m"      'add-mode-abbrev)
(define-key my-abbrev-map "\C-m"   'inverse-add-mode-abbrev)
(define-key my-abbrev-map "r"      'expand-region-abbrevs)
(define-key my-abbrev-map "u"      'unexpand-abbrev)
(define-key my-abbrev-map "v"      'edit-abbrevs)
(define-key my-abbrev-map "x"      'dabbrev-expand)
(define-key my-abbrev-map " "      'set-abbrev-mark)

(when (featurep 'icicles)
  (define-key my-abbrev-map [escape] (make-sparse-keymap)) ; to allow icicle-complete-keys
  (add-to-list 'icicle-keymaps-for-key-completion 'my-abbrev-map t))

    ;; Org-Mode

(define-prefix-command 'my-org-prefix 'my-org-map "Org Mode")
(when (featurep 'org)
  (define-key my-org-map "a" 'org-agenda)
  (define-key my-org-map "l" 'org-store-link))

(when (featurep 'icicles)
  (define-key my-org-map [escape] (make-sparse-keymap)) ; to allow icicle-complete-keys
  (add-to-list 'icicle-keymaps-for-key-completion 'my-org-map t))

    ;; Icicles

(define-prefix-command 'my-icicles-prefix 'my-icicles-map "Icicles")
(when (featurep 'icicles)
  (define-key my-icicles-map "b" 'icicle-search-buffer)
  (define-key my-icicles-map "f" 'icicle-search-file)
  (define-key my-icicles-map "o" 'icicle-occur)
  (define-key my-icicles-map "m" 'icicle-search-bookmark)
  (define-key my-icicles-map "s" 'icicle-search)
  (define-key my-icicles-map "t" 'icicle-search-bookmarks-together)
  (define-key my-icicles-map "$" 'icicle-search-word)
  (define-key my-icicles-map "^" 'icicle-search-keywords)
  (define-key my-icicles-map "`" 'icicle-compilation-search)
  (define-key my-icicles-map "=" 'icicle-imenu)
  (define-key my-icicles-map [?\"] 'icicle-search-char-property) 
  (define-key my-icicles-map [tab] 'icicle-comint-command)
  (define-key my-icicles-map "c" 'icicle-imenu-command)
  (define-key my-icicles-map [escape] (make-sparse-keymap)) ; to allow icicle-complete-keys
  (add-to-list 'icicle-keymaps-for-key-completion 'my-icicles-map t))


    ;; Anything

(define-prefix-command 'my-anything-prefix 'my-anything-map "Anything Interface")
(when (featurep 'anything)
  (define-key my-anything-map "a" 'anything)
  (define-key my-anything-map "b" 'anything-bookmarks)
  (define-key my-anything-map "c" 'anything-browse-code))

(when (featurep 'icicles)
  (define-key my-anything-map [escape] (make-sparse-keymap)) ; to allow icicle-complete-keys
  (add-to-list 'icicle-keymaps-for-key-completion 'my-anything-map t))

    ;; Links

(define-prefix-command 'my-links-prefix 'my-links-map "Links")
(when (featurep 'org)
  (define-key my-links-map "o" 'org-store-link))
(when (featurep 'linkd)
  (define-key my-links-map "b" 'linkd-block-around-point)
  (define-key my-links-map "e" 'linkd-edit-link-at-point)
  (define-key my-links-map "i" 'linkd-insert-link)
  (define-key my-links-map "p" 'linkd-process-block)
  (define-key my-links-map "s" 'linkd-send-block-to-shell)
  (define-key my-links-map [return] 'linkd-follow-at-point)
  (define-key my-links-map "\C-n" 'linkd-next-link)
  (define-key my-links-map "\C-p" 'linkd-previous-link))

(when (featurep 'icicles)
  (define-key my-links-map [escape] (make-sparse-keymap)) ; to allow icicle-complete-keys
  (add-to-list 'icicle-keymaps-for-key-completion 'my-links-map t))


;;
;; Keybinding Setters
;;

(defun my-set-global-sysdep-keybindings ()
  "Sets global keybindings that are system
   dependent (operating system or emacs version)."
   nil
  )

(defun my-set-global-keybindings ()
  "Adjusts default global keybindings to
   conform to my preferences."
  ;;
  ;; Movement
  ;;
    (global-set-key "\C-a"        'my-move-beginning-of-line)
    (global-set-key "\C-w"        'scroll-down-or-beg)
    (global-set-key "\C-v"        'scroll-up-or-end)
    (global-set-key "\M-v"        'end-of-buffer)
    (global-set-key "\M-w"        'beginning-of-buffer)
    (global-set-key "\M-\C-v"     'scroll-other-window)
    (global-set-key "\M-\C-w"     'scroll-other-window-down)
    (global-set-key "\M-p"        'backward-up-list)
    (global-set-key "\M-n"        'up-list)
    (global-set-key "\M-\C-f"     'forward-sexp-or-char)
    (global-set-key "\M-\C-b"     'backward-sexp-or-char)
    (global-set-key [?\A-a]       'back-to-indentation)
  ;;
  ;; Basic Functions
  ;;
    (global-set-key "\C-x\C-m"    'execute-extended-command)
    (global-set-key "\C-c\C-m"    'execute-extended-command)
    (when (featurep 'lacarte)
      (global-set-key "\C-x\M-m"  'lacarte-execute-command)
      (global-set-key "\C-c\M-m"  'lacarte-execute-command))
    (global-set-key [?\C-0]       'universal-argument)
  ;;
  ;; Help
  ;;   I've moved the help command to enable more efficient
  ;;   access to some editing commands. But this necessitates
  ;;   a bit of extra work to ensure consistency elswhere,
  ;;   especially isearch and icicles (see below for the former)
  ;;   and (@> "File and Navigation Modes") for the latter.
  ;;
    (define-key help-map "\C-c"   'describe-key-briefly)
    (define-key help-map "\M-c"   'describe-copying)
    (global-set-key "\C-c\C-h"    'help-command)
    (global-set-key "\M-?"        'help-command)
    (setq help-event-list (cons [?\C-c ?\C-h] (cons [?\M-\?] help-event-list)))
  ;;
  ;; Searching
  ;;
    (global-set-key "\M-s"    'isearch-forward-regexp)
    (global-set-key "\M-r"    'isearch-backward-regexp)
    (global-set-key "\C-\M-s" 'query-replace)
    (global-set-key "\C-\M-r" 'query-replace-regexp)
    (global-set-key [?\C-5]   'query-replace)
    (global-set-key [?\C-6]   'query-replace-regexp)
    (defalias 'qr  'query-replace)
    (defalias 'qrr 'query-replace-regexp)
    ;; To end a search with a specially defined key, like a keypad key
    ;; and to have the search not automatically continue with a single
    ;; C-s, do the following for the key.  For example, this makes
    ;; kp-2 act like a proper search ender.
    ;; (define-key isearch-mode-map [kp-2] 'isearch-other-control-char)
    (define-key isearch-mode-map "\M-w" 'isearch-yank-word-or-char)
    (define-key isearch-mode-map "\C-w" 'isearch-other-control-char)
    (define-key isearch-mode-map [(control return)] 'isearch-exit)
    ;(define-key isearch-mode-map "\C-c\C-h" (lookup-key isearch-mode-map "\C-h"))
    (define-key isearch-mode-map "\M-?" (lookup-key isearch-mode-map "\C-h"))
    (define-key isearch-mode-map "\C-h" 'isearch-delete-char)
    (define-key minibuffer-local-isearch-map "\M-s" 'isearch-forward-exit-minibuffer)
    (define-key minibuffer-local-isearch-map "\M-r" 'isearch-reverse-exit-minibuffer)
    (define-key minibuffer-local-isearch-map [(control meta tab)] 'isearch-complete-edit)
    (define-key minibuffer-local-isearch-map [(meta control ?i)]  nil)
  ;;
  ;; Kills and Copies
  ;;
    (global-set-key "\C-k"            'kill-line)
    (global-set-key "\M-k"            'kill-region)
    (global-set-key "\M-\C-k"         'kill-sexp)
    (global-set-key [?\S-\C-k]        'kill-sentence)
    (global-set-key [?\S-\M-\C-k]     'kill-paragraph)
    (global-set-key [?\A-k]           'kill-whole-line)
    (global-set-key [?\M-B]           'backward-kill-word)
    (global-set-key [?\A-b]           'backward-kill-word)
    (global-set-key [\M-backspace]    'backward-kill-word)
    (global-set-key [?\A-K]           'backward-kill-sentence)
    (global-set-key [?\S-\C-\A-k]     'backward-kill-paragraph)
    (global-set-key [?\C-\A-k]        'backward-kill-sexp)
    (global-set-key [\C-\S-backspace] 'backward-kill-sentence)
    (global-set-key [\C-\M-backspace] 'backward-kill-sexp)
    (global-set-key "\M-c"            'my-kill-ring-save) ;alternate: my-copy-region-as-kill
    (global-set-key "\M-\C-y"         'append-next-kill)
    (global-set-key "\C-z"            'zap-to-char)
    (global-set-key "\M-z"            'undo)
    (global-set-key "\C-h"        'delete-backward-char) ; Move help command
    (global-set-key "\M-h"        'backward-kill-word)
  ;;
  ;; Marking
  ;;
    (global-set-key [?\M- ]       'mark-sexp)
    (global-set-key [?\C-\M- ]    'mark-defun)
    (global-set-key "\M-m"        'mark-paragraph)
    (global-set-key "\C-\M-m"     'mark-end-of-sentence)
    (global-set-key [?\S-\C-\M-m] 'mark-page)
    (global-set-key [?\A-m]       'mark-whole-buffer)
    (global-set-key [?\A-M]       'mark-word)
  ;;
  ;; Buffers, Files, Windows, and Frames
  ;;
    (global-set-key "\C-x\C-b"    'ibuffer-list-buffers-select)
    (global-set-key [?\C-x ?\C-\S-b] 'ibuffer-list-buffers) ; no select and files arg
    (global-set-key "\C-xy"       'bury-buffer)
    (global-set-key "\C-xo"       'other-window)         ;; also bound to C-x C-o
    (global-set-key "\C-x\C-o"    'other-window)         ;; same to avoid collisions
    (global-set-key "\C-x\M-w"    'append-to-file)
    (global-set-key "\C-x\M-\C-f" 'find-file-at-point)
    (global-set-key "\C-x:"       'view-url-in-buffer)
    (global-set-key "\C-\M-z"     'iconify-or-deiconify-frame)
  ;;
  ;; Positioning
  ;;
    (global-set-key "\M-l"        'reposition-window)
    (global-set-key "\M-\C-l"     'move-to-window-line)
    (global-set-key "\C-xL"       'count-lines-page) ;; ATTN replace with something smarter
    (global-set-key "\C-xl"       'what-line)
    (global-set-key "\C-xg"       'goto-line)    ;; ATTN: used to be C-xt
    (global-set-key "\M-g"        'my-go-prefix) ;; replaces similar map
  ;;
  ;; Macros   
  ;;
    (global-set-key "\C-x\M-k"    'kmacro-keymap)
    (global-set-key [?\C-3]       'kmacro-start-macro-or-insert-counter)
    (global-set-key [?\C-4]       'kmacro-end-or-call-macro)
  ;;
  ;; Miscellaneous Operations
  ;;
    (global-set-key [?\C-7]       'align-regexp)  ;; very useful! mnemnoic: & for align
    (global-set-key [?\A-q]       'align-regexp)  
    (global-set-key "\C-xq"       'align-regexp)  
    (global-set-key [\A-return]   'delete-blank-lines)
    (global-set-key [?\A- ]       'just-one-space)
    (global-set-key [?\C-\A- ]    'delete-trailing-whitespace)
    (global-set-key "\C-\\"       'comment-dwim)
    (global-set-key "\M-/"        'hippie-expand)
  ;;
  ;; Rectangle and Register Commands
  ;;
    (global-set-key "\C-xra"      'append-to-register)
    (global-set-key "\C-xrc"      'copy-to-register)            
    (global-set-key "\C-xrl"      'list-registers)            
    (global-set-key "\C-xrp"      'prepend-to-register)
    (global-set-key "\C-xrs"      'string-rectangle)
    (global-set-key "\C-xrt"      'register-to-point)
    (global-set-key "\C-xrv"      'view-register)
    (global-set-key "\C-xrx"      'clear-rectangle)
    (global-set-key "\C-xr "      'point-to-register)
  ;;
  ;; Attach additional keymaps and related rebindings
  ;;
    (define-key ctl-x-map "t"  'my-tags-prefix)  ;; ATTN: used to be C-xg (see goto-line above)
    (define-key ctl-x-map "%"  'my-string-replace-prefix)
    (define-key ctl-x-map "m"  'my-mark-prefix)
    (define-key ctl-x-map "a"  'my-abbrev-prefix)
    (global-set-key [?\A-t]    'my-transpose-prefix)
    (global-set-key [?\S-\M-t] 'my-transpose-prefix)
  ;;
  ;; Modified Arrow Keys get added functionality
  ;;
    (global-set-key [C-left]      'back-to-indentation)
    (global-set-key [C-right]     'end-of-line)
    (global-set-key [M-left]      'backward-word)
    (global-set-key [M-right]     'forward-word)
    (global-set-key [C-M-left]    'backward-list)
    (global-set-key [C-M-right]   'forward-list)
    (global-set-key [C-up]        'scroll-down-or-beg)
    (global-set-key [C-down]      'scroll-up-or-end)
    (global-set-key [M-up]        'backward-paragraph)
    (global-set-key [M-down]      'forward-paragraph)
    (global-set-key [C-M-up]      'backward-page)
    (global-set-key [C-M-down]    'forward-page)
  ;;
  ;; System Dependent Global Bindings
  ;;
    (my-set-global-sysdep-keybindings)
  ;;
  ;; Third-Party Mode Global Access Keys
  ;;
    (global-set-key "\C-ca"       'my-anything-prefix)
    (global-set-key "\C-ci"       'my-icicle-prefix)
    (global-set-key "\C-cl"       'my-links-prefix)
    (global-set-key "\C-co"       'my-org-prefix)
    (when (featurep 'icicles)
      (global-set-key "\C-x\M-\C-o" 'icicle-switch-to/from-minibuffer))
    (when (featurep 'lacarte)
      (global-set-key [?\e ?\M-x]   'lacarte-execute-command)
      (global-set-key [?\M-`]       'lacarte-execute-command))
)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                                                           ;;
;;(@* "Settings and Initialization")                         ;;
;;                                                           ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Start-up configuration

(setq inhibit-startup-message 't)
(setq initial-major-mode 'lisp-interaction-mode) ; formerly fundamental-mode
(setq initial-frame-alist my-frame-alist)        ; see (@> "Environment Configuration")

  ; load custom-file in (@> "Initial State")
(setq custom-file (expand-file-name "~/.emacs.d/emacs-custom.el")) 

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Enable Disabled Commands

(put 'eval-expression  'disabled nil)
(put 'narrow-to-region 'disabled nil)
(put 'downcase-region  'disabled nil)
(put 'upcase-region    'disabled nil)
(put 'set-goal-column  'disabled nil)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Basic Operation

(defun my-set-completion-mode (sym)
  "Choose among various intelligent completion methods
   according to symbol. Current choices are 'icicles, 'ido,
   'iswitchb, 'none."
   (cond
    ((equal sym 'icicles)
     (if (not (featurep 'icicles))
         (require-soft 'icicles))
     (if (featurep 'icicles)
         (icicle-mode 1)
       (message "Cannot load icicles")))
    ((equal sym 'ido)
     (ido-mode '1)
     (setq ido-case-fold t))
    ((equal sym 'iswitchb) ;;; Use smart buffer switching
     (iswitchb-mode '1)                      
     (setq iswitchb-case t))
    ))
   
    
(defun my-set-operating-state ()
  "Set Emacs to preferred modes, maps, and more."
  (setq-default case-fold-search nil)
  (setq-default indent-tabs-mode nil)
  (setq-default next-line-add-newlines nil)
    
  (setq-default fill-column '72)
  (setq paragraph-start paragraph-separate)            ;;; Use blank lines to separate paragraphs by default
  (setq adaptive-fill-regexp "[ \t]*\\([>*%#]+ +\\)?") ;;; Fill around comment beginnings and yanked-messages
  (setq sentence-end-double-space nil)                 ;;; Allow frenchspacing
  (setq page-delimiter "^\\(\f\\|\n\n+\\)")            ;;; FF or 2+ consecutive blank lines

  (setq history-length 256)
  (setq print-length 1024)                 ;;; Give more information in help
  (setq print-level  8)

  (if (>= emacs-major-version 23)
      (setq-default major-mode 'org-mode)
    (setq default-major-mode 'org-mode))   ;; back to fundamental-mode if org-mode causes strange behavior??
  (setq my-keep-scratch-buf "*elisp*")     ;; if nil, delete; if string, new name; otherwise leave alone.
  
  (setq display-time-24hr-format t)
  (display-time)                           ;;; Time on Mode Line
    
  (transient-mark-mode '1)                 ;;; Highlight region
  (setq kill-read-only-ok t)               ;;; OK to use kill to copy text in read-only buffer
  (if (fboundp 'menu-bar-mode)
      (menu-bar-mode  1))                  ;;; Make menu bar available, just in case
  (if (fboundp 'tool-bar-mode)
      (tool-bar-mode  -1))                 ;;; Disable tool bar
  (setq-default scroll-bar-mode 'right)    ;;; Put scroll bars on the right
  (toggle-scroll-bar  '1)                  ;;; Use scroll bars
  (auto-compression-mode '1)               ;;; Auto view compressed files
  (turn-on-global-show-paren-mode)         ;;; show matching parens
  (setq indicate-empty-lines t)            ;;; show empty lines at end of file

  (setq browse-url-browser-function 'browse-url-firefox) ;; use new tabs in firefox
  (setq browse-url-firefox-new-window-is-tab t)
  (setq browse-url-new-window-flag  t)

  (my-set-completion-mode 'icicles)
  (my-set-global-keybindings)
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                                                           ;;
;;(@* "System Dependent Settings")                           ;;
;;                                                           ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(let* ((ostype (getenv "OSTYPE"))
       (myos (and ostype (downcase (getenv "OSTYPE")))))
  (when myos
    (cond
     ((string-equal myos "linux")
      (setq list-directory-verbose-switches "-HlrtF")
      (setq list-directory-brief-switches   "-CHvF"))
     (t
      (setq list-directory-verbose-switches "-lrtF")
      (setq list-directory-brief-switches   "-CF"))
     )))

(when window-system
  (setq frame-title-format '("" invocation-name " <%b>"))
  (tooltip-mode -1)
  (mouse-wheel-mode t)
  (blink-cursor-mode t)
  (setq visible-bell 't)           
  )

(when (eq window-system 'x)
  (set-face-background 'highlight "tan")   ;;; "red" is nice also
  )

;; Emacs Version Specific Settings

(when (equal emacs-major-version 23)
  (set-cursor-color "gray40")
  )


;; Mac OSX Specific Settings

(when (eq window-system 'mac)
  (message "MAC CARBON SYSTEM CHECK")) ; checking if this is sufficient

(when (eq window-system 'ns)
  (message "MAC COCOA SYSTEM CHECK")) ; checking if this is sufficient

(when (or (eq window-system 'ns)
          (eq window-system 'mac))
  (setq mac-command-modifier  'meta) ;;; Emacs 23 defaults are different
  (setq mac-option-modifier   'alt)
  (setq mac-function-modifier 'super)
  (setq mac-emulate-three-button-mouse 'reverse)
  (setq mac-pass-command-to-system nil)  ; not sure about the last two
  (setq mac-pass-control-to-system nil))
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                                                           ;;
;;(@* "Faces")                                               ;;;;;;
;;                                                           ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(if window-system
    (progn
      (set-face-background 'region "lightgoldenrod2")  
      (set-face-background 'secondary-selection "gray")
      )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                                                           ;;
;;(@* "Initial State")                                       ;;
;;                                                           ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(load-file custom-file)   ;; holds customization-only customizations
(my-set-operating-state)

;; handle scratch buffer if requested
(my-aif (get-buffer "*scratch*")  
    (cond
     ((stringp my-keep-scratch-buf)
      (with-current-buffer it
        (rename-buffer my-keep-scratch-buf)
        (funcall initial-major-mode)))
     ((null my-keep-scratch-buf)
      (kill-buffer it))))

;; load additional environment variables, if any
;; Note: Mac OS X uses ~/.MacOSX/environment.plist
;; Might also want some emacs-specific environment settings.
(let ((envir (concat my-home-lisp-dir "my-env.el")))
  (if (file-exists-p envir)
    (load-file envir)))  

;; start in the shell
(shell)
